//////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2007 Xilinx, Inc.
// This design is confidential and proprietary of Xilinx, All Rights Reserved.
//////////////////////////////////////////////////////////////////////////////
//   ____  ____
//  /   /\/   /
// /___/  \  /   Vendor:        Xilinx
// \   \   \/    Version:       1.0.0
//  \   \        Filename:      framebuf.v
//  /   /        Date Created:  July 1, 2007
// /___/   /\    Last Modified: July 1, 2007
// \   \  /  \
//  \___\/\___\
//
// Devices:   Spartan-3 Generation FPGA
// Purpose:   Frame Buffer design top level
// Contact:   crabill@xilinx.com
// Reference: None
//
// Revision History:
//   Rev 1.0.0 - (crabill) Created July 1, 2007 for PCB revision D.
//
//////////////////////////////////////////////////////////////////////////////
//
// LIMITED WARRANTY AND DISCLAIMER. These designs are provided to you "as is".
// Xilinx and its licensors make and you receive no warranties or conditions,
// express, implied, statutory or otherwise, and Xilinx specifically disclaims
// any implied warranties of merchantability, non-infringement, or fitness for
// a particular purpose. Xilinx does not warrant that the functions contained
// in these designs will meet your requirements, or that the operation of
// these designs will be uninterrupted or error free, or that defects in the
// designs will be corrected. Furthermore, Xilinx does not warrant or make any
// representations regarding use or the results of the use of the designs in
// terms of correctness, accuracy, reliability, or otherwise.
//
// LIMITATION OF LIABILITY. In no event will Xilinx or its licensors be liable
// for any loss of data, lost profits, cost or procurement of substitute goods
// or services, or for any special, incidental, consequential, or indirect
// damages arising from the use or operation of the designs or accompanying
// documentation, however caused and on any theory of liability. This
// limitation will apply even if Xilinx has been advised of the possibility
// of such damage. This limitation shall apply not-withstanding the failure
// of the essential purpose of any limited remedies herein.
//
//////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2007 Xilinx, Inc.
// This design is confidential and proprietary of Xilinx, All Rights Reserved.
//////////////////////////////////////////////////////////////////////////////

`timescale 1 ns / 1 ps

module framebuf
  (
  input  wire        CLK_TST,

  output reg         LCD_E,
  output reg         LCD_RS,
  output reg         LCD_RW,
  output reg   [7:0] LCD_DB,
  output reg   [7:0] LED,
  output reg         CRCERR,

  input  wire        RS232_DCE_RXD,
  output wire        RS232_DCE_TXD,

  inout  wire        REG1_SDA,
  inout  wire        REG1_SCL,
  inout  wire        REG2_SDA,
  inout  wire        REG2_SCL,

  input  wire        BTN_EAST,
  input  wire        BTN_NORTH,
  input  wire        BTN_SOUTH,
  input  wire        BTN_WEST,
  input  wire  [3:0] SW,
  input  wire        ROT_A,
  input  wire        ROT_B,
  input  wire        ROT_CENTER,

  output reg   [3:0] VGA_R,
  output reg   [3:0] VGA_G,
  output reg   [3:0] VGA_B,
  output reg         VGA_HSYNC,
  output reg         VGA_VSYNC,

  output wire [21:0] NF_A,
  input  wire  [7:0] NF_D,
  output wire        NF_BYTE,
  output wire        NF_RP,
  output wire        NF_WP,
  output wire        NF_WE,
  output wire        NF_CE,
  output wire        NF_OE,
  output wire        FPGA_INIT_B,
  output wire        SPI_SS_B,
  output wire        ALT_SS_B,

  output wire [12:0] SD_A,
  output wire  [1:0] SD_BA,
  output wire        SD_CAS,
  output wire        SD_CK_P,
  output wire        SD_CK_N,
  output wire        SD_CKE,
  output wire        SD_CS,
  inout  wire [15:0] SD_DQ,
  output wire        SD_LDM,
  output wire        SD_RAS,
  output wire        SD_UDM,
  output wire        SD_ODT,
  output wire        SD_UDQS_P,
  output wire        SD_UDQS_N,
  output wire        SD_LDQS_P,
  output wire        SD_LDQS_N,
  output wire        SD_WE,
  output wire        SD_LOOP_OUT,
  input  wire        SD_LOOP_IN
  );

  //******************************************************************//
  // Create global clocks from external signals.  This method for the //
  // generation of the 200 MHz clock is not recommended; ideally, you //
  // would be using a direct 200 MHz, low-jitter clock source.  This  //
  // arrangement is good for a self-contained demo.                   //
  //******************************************************************//

  wire          bufged_aux; // memory and pixel clock
  wire          bufged_a90; // memory clock shifted
  wire          bufged_div; // derived picoblaze clock
  wire          nubpre1x;   //  50 MHz with 50 MHz input
  wire          nubpre2x;   // 100 MHz with 50 MHz input
  wire          nubpre4x;   // 200 MHz with 50 MHz input
  wire          clocklock;
  wire          predelay0;
  wire          predelay1;
  wire          prelocked;

  DCM_SP #(
    .DESKEW_ADJUST("SOURCE_SYNCHRONOUS"),
    .CLK_FEEDBACK("2X"),
    .CLKFX_DIVIDE(1),
    .CLKFX_MULTIPLY(4),
    .CLKIN_PERIOD(20.000))
  DCM_SP_INST (
    .CLKIN(CLK_TST),
    .CLKFB(nubpre2x),
    .RST(1'b0),
    .PSEN(1'b0),
    .PSINCDEC(1'b0),
    .PSCLK(1'b0),
    .DSSEN(1'b0),
    .CLK0(nubpre1x),
    .CLK90(),
    .CLK180(),
    .CLK270(),
    .CLKDV(),
    .CLK2X(nubpre2x),
    .CLK2X180(),
    .CLKFX(nubpre4x),
    .CLKFX180(),
    .STATUS(),
    .LOCKED(predelay0),
    .PSDONE());

  synchro #(.INITIALIZE("LOGIC0"))
  synchro_lock0 (.async(predelay0),.sync(predelay1),.clk(nubpre1x));
  synchro #(.INITIALIZE("LOGIC0"))
  synchro_lock1 (.async(predelay1),.sync(prelocked),.clk(nubpre1x));

  //******************************************************************//
  // Instantiate PicoBlaze and the Instruction ROM.                   // 
  //******************************************************************//

  wire    [9:0] address;
  wire   [17:0] instruction;
  wire    [7:0] port_id;
  wire    [7:0] out_port;
  reg     [7:0] in_port;
  wire          write_strobe;
  wire          read_strobe;
  wire          reset;
  reg           interrupt;
  wire          interrupt_ack;
  wire          interrupt_request;

  kcpsm3 kcpsm3_inst (
    .address(address),
    .instruction(instruction),
    .port_id(port_id),
    .write_strobe(write_strobe),
    .out_port(out_port),
    .read_strobe(read_strobe),
    .in_port(in_port),
    .interrupt(interrupt),
    .interrupt_ack(interrupt_ack),
    .reset(reset),
    .clk(bufged_div));

  control control_inst (
    .address(address),
    .instruction(instruction),
    .clk(bufged_div));

  wire          interrupt_clear;

  always @(posedge bufged_div)
  begin
    if (interrupt_request) interrupt <= 1'b1;
    else if (interrupt_clear) interrupt <= 1'b0;
  end

  assign interrupt_clear = interrupt_ack || (write_strobe && (port_id == 8'hff));

  //******************************************************************//
  // Generate synchronous one-shot reset.                             // 
  //******************************************************************//

  synchro #(.INITIALIZE("LOGIC1"))
  synchro_reset (.async(!clocklock),.sync(reset),.clk(bufged_div));

  //******************************************************************//
  // Self-contained stub for voltage regulator control.               // 
  //******************************************************************//

  rs232reg rs232reg_inst (
    .clk(bufged_div),
    .reset(reset),
    .rs232_dce_rxd(RS232_DCE_RXD),
    .rs232_dce_txd(RS232_DCE_TXD),
    .reg1_sda(REG1_SDA),
    .reg1_scl(REG1_SCL),
    .reg2_sda(REG2_SDA),
    .reg2_scl(REG2_SCL));

  //******************************************************************//
  // Instantiate the button processor.                                //
  // Port 00h is read only; switch status [3:0]                       //
  // Port 01h is read only; button status [3:0]                       //
  // Port 02h is read for knob status [3:0], write clears             //
  //******************************************************************//

  wire  [3:0] btn_out;
  wire  [3:0] sws_out;
  wire  [3:0] rot_out;
  wire        rot_clr;

  picouser picouser_inst (
    .BTN_EAST(BTN_EAST),
    .BTN_NORTH(BTN_NORTH),
    .BTN_SOUTH(BTN_SOUTH),
    .BTN_WEST(BTN_WEST),
    .SW(SW),
    .ROT_A(ROT_A),
    .ROT_B(ROT_B),
    .ROT_CENTER(ROT_CENTER),
    .rot_clr(rot_clr),
    .clk(bufged_div),
    .btn_out(btn_out),
    .sws_out(sws_out),
    .rot_out(rot_out));

  assign rot_clr = write_strobe && (port_id == 8'h02);

  //******************************************************************//
  // Implement the LCD/LED interface.                                 //
  // Port 03h is read/write LCD data out                              //
  // Port 04h is read/write LCD control out                           //
  //          bit 2 -- lcd rs                                         //
  //          bit 1 -- lcd rw                                         //
  //          bit 0 -- lcd e                                          //
  //          others not implemented                                  //
  // Port 05h is read/write LED data out                              //
  //******************************************************************//

  always @(posedge bufged_div)
  begin
    if (write_strobe)
    begin
      if (port_id == 8'h03) {LCD_DB} <= out_port;
      if (port_id == 8'h04) {LCD_RS,LCD_RW,LCD_E} <= out_port[2:0];
      if (port_id == 8'h05) LED <= out_port;
      if (port_id == 8'hf0) CRCERR <= out_port[0];
    end
  end

  //******************************************************************//
  // Instantiate a programmable video timing controller.              //
  // This uses ports 06h through 17h, all as read/write.              //
  //******************************************************************//

  reg  [11:0] tc_hsblnk;
  reg  [11:0] tc_hssync;
  reg  [11:0] tc_hesync;
  reg  [11:0] tc_heblnk;

  reg  [11:0] tc_vsblnk;
  reg  [11:0] tc_vssync;
  reg  [11:0] tc_vesync;
  reg  [11:0] tc_veblnk;

  reg         tc_vs_neg;
  reg         tc_hs_neg;

  always @(posedge bufged_div)
  begin
    if (write_strobe)
    begin
      if (port_id == 8'h07) {tc_hs_neg,tc_vs_neg} <= out_port[1:0];
      if (port_id == 8'h08) tc_hsblnk[7:0] <= out_port;
      if (port_id == 8'h09) tc_hsblnk[11:8] <= out_port[3:0];
      if (port_id == 8'h0a) tc_hssync[7:0] <= out_port;
      if (port_id == 8'h0b) tc_hssync[11:8] <= out_port[3:0];
      if (port_id == 8'h0c) tc_hesync[7:0] <= out_port;
      if (port_id == 8'h0d) tc_hesync[11:8] <= out_port[3:0];
      if (port_id == 8'h0e) tc_heblnk[7:0] <= out_port;
      if (port_id == 8'h0f) tc_heblnk[11:8] <= out_port[3:0];
      if (port_id == 8'h10) tc_vsblnk[7:0] <= out_port;
      if (port_id == 8'h11) tc_vsblnk[11:8] <= out_port[3:0];
      if (port_id == 8'h12) tc_vssync[7:0] <= out_port;
      if (port_id == 8'h13) tc_vssync[11:8] <= out_port[3:0];
      if (port_id == 8'h14) tc_vesync[7:0] <= out_port;
      if (port_id == 8'h15) tc_vesync[11:8] <= out_port[3:0];
      if (port_id == 8'h16) tc_veblnk[7:0] <= out_port;
      if (port_id == 8'h17) tc_veblnk[11:8] <= out_port[3:0];
    end
  end

  wire   [11:0] hcount;
  wire          hsync;
  wire          hblnk;

  wire   [11:0] vfetch;
  wire          active;

  wire   [11:0] vcount;
  wire          vsync;
  wire          vblnk;

  fbctime fbctime_inst (
    .tc_hsblnk(tc_hsblnk),
    .tc_hssync(tc_hssync),
    .tc_hesync(tc_hesync),
    .tc_heblnk(tc_heblnk),
    .hcount(hcount),
    .hsync(hsync),
    .hblnk(hblnk),
    .tc_vsblnk(tc_vsblnk),
    .tc_vssync(tc_vssync),
    .tc_vesync(tc_vesync),
    .tc_veblnk(tc_veblnk),
    .vfetch(vfetch),
    .active(active),
    .vcount(vcount),
    .vsync(vsync),
    .vblnk(vblnk),
    .clk(bufged_aux));

  // I want a vertical blanking interrupt, but
  // first the vblank needs to be synchronized
  // to the processor clock.

  wire          resync_vblnk;
  reg           resync_vblnk_delay;

  synchro #(.INITIALIZE("LOGIC0"))
  synchro_vbint (.async(vblnk),.sync(resync_vblnk),.clk(bufged_div));

  always @(posedge bufged_div) resync_vblnk_delay <= resync_vblnk;
  assign interrupt_request = resync_vblnk & !resync_vblnk_delay;

  //******************************************************************//
  // Instantiate two line buffers and the output logic.               //
  //******************************************************************//

  reg    [11:0] hcount_frm_pipe;
  reg    [11:0] vcount_frm_pipe;

  always @(posedge bufged_aux) hcount_frm_pipe <= hcount;
  always @(posedge bufged_aux) vcount_frm_pipe <= vcount;

  wire   [31:0] memorydata;
  wire    [9:0] memoryaddr;
  wire          memorywrite0;
  wire          memorywrite1;

  wire   [15:0] pixeldata0;
  wire   [15:0] pixeldata1;
  
  linebuf linebuf0_inst (
    .hcount(hcount_frm_pipe[10:0]),
    .pixeldata(pixeldata0),
    .readenable(!vcount_frm_pipe[0]),
    .memorydata(memorydata),
    .memoryaddr(memoryaddr),
    .memorywrite(memorywrite0),
    .clk(bufged_aux));

  linebuf linebuf1_inst (
    .hcount(hcount_frm_pipe[10:0]),
    .pixeldata(pixeldata1),
    .readenable(vcount_frm_pipe[0]),
    .memorydata(memorydata),
    .memoryaddr(memoryaddr),
    .memorywrite(memorywrite1),
    .clk(bufged_aux));

  //******************************************************************//
  // This is the output logic.                                        //
  //******************************************************************//

  reg     [5:0] hsync_delayed;
  reg     [5:0] vsync_delayed;
  reg     [5:0] pixeldata_sel;
  reg     [5:0] pixelblanking;
  reg     [5:0] crc_vsync_rst;

  always @(posedge bufged_aux)
  begin
    pixeldata_sel[5:0] <= {pixeldata_sel[4:0], vcount[0]};
    pixelblanking[5:0] <= {pixelblanking[4:0], (hblnk || vblnk)};
    hsync_delayed[5:0] <= {hsync_delayed[4:0], (hsync ^ tc_hs_neg)};
    vsync_delayed[5:0] <= {vsync_delayed[4:0], (vsync ^ tc_vs_neg)};
    crc_vsync_rst[5:0] <= {crc_vsync_rst[4:0], vsync};
    VGA_HSYNC <= hsync_delayed[5];
    VGA_VSYNC <= vsync_delayed[5];
  end

  // Continue with the output pipeline,
  // implement the data select mux and
  // the fade in, fade out logic.

  reg     [3:0] crc_r_d3;
  reg     [3:0] crc_g_d3;
  reg     [3:0] crc_b_d3;
  reg     [3:0] crc_x_d3;
  reg           crc_dval;

  reg     [3:0] frame_r_d3;
  reg     [3:0] frame_g_d3;
  reg     [3:0] frame_b_d3;

  reg     [3:0] frame_r_d4;
  reg     [3:0] frame_g_d4;
  reg     [3:0] frame_b_d4;

  reg     [3:0] fade_mult;
  reg     [2:0] disp_fade;
  wire    [7:0] multb;
  wire    [7:0] multg;
  wire    [7:0] multr;

  assign multb = fade_mult * frame_b_d3;
  assign multg = fade_mult * frame_g_d3;
  assign multr = fade_mult * frame_r_d3;

  always @(posedge bufged_aux)
  begin
    crc_x_d3 <= pixeldata_sel[2] ? pixeldata1[15:12] : pixeldata0[15:12];
    crc_b_d3 <= pixeldata_sel[2] ? pixeldata1[11: 8] : pixeldata0[11: 8];
    crc_g_d3 <= pixeldata_sel[2] ? pixeldata1[ 7: 4] : pixeldata0[ 7: 4];
    crc_r_d3 <= pixeldata_sel[2] ? pixeldata1[ 3: 0] : pixeldata0[ 3: 0];
    crc_dval <= !pixelblanking[2];

    frame_b_d3 <= pixeldata_sel[2] ? pixeldata1[11: 8] : pixeldata0[11: 8];
    frame_g_d3 <= pixeldata_sel[2] ? pixeldata1[ 7: 4] : pixeldata0[ 7: 4];
    frame_r_d3 <= pixeldata_sel[2] ? pixeldata1[ 3: 0] : pixeldata0[ 3: 0];

    // frame_b_d4 <= frame_b_d3 >> disp_fade;
    // frame_g_d4 <= frame_g_d3 >> disp_fade;
    // frame_r_d4 <= frame_r_d3 >> disp_fade;

    frame_b_d4 <= multb[6:3];
    frame_g_d4 <= multg[6:3];
    frame_r_d4 <= multr[6:3];
  end

  // These signals from the hardware cursor
  // logic, arrive after five cycles of delay
  // and would be combined with signals that
  // have a delay index of [4].  Here, I am
  // making the first simplification which is
  // to require that hardware cursors never
  // overlap -- this allows a simple method
  // to sum all of the cursors.

  reg     [3:0] cursor0_r;
  reg     [3:0] cursor0_g;
  reg     [3:0] cursor0_b;
  reg           cursor0_t;

  reg     [3:0] cursor1_r;
  reg     [3:0] cursor1_g;
  reg     [3:0] cursor1_b;
  reg           cursor1_t;

  reg     [3:0] cursor2_r;
  reg     [3:0] cursor2_g;
  reg     [3:0] cursor2_b;
  reg           cursor2_t;

  reg     [3:0] cursor3_r;
  reg     [3:0] cursor3_g;
  reg     [3:0] cursor3_b;
  reg           cursor3_t;

  reg     [3:0] cursor4_r;
  reg     [3:0] cursor4_g;
  reg     [3:0] cursor4_b;
  reg           cursor4_t;

  reg     [3:0] cursor5_r;
  reg     [3:0] cursor5_g;
  reg     [3:0] cursor5_b;
  reg           cursor5_t;

  wire    [3:0] hwc_sum_r_d4;
  wire    [3:0] hwc_sum_g_d4;
  wire    [3:0] hwc_sum_b_d4;
  wire          hwc_sum_t_d4;

  assign hwc_sum_r_d4 = cursor0_r | cursor1_r | cursor2_r |
                        cursor3_r | cursor4_r | cursor5_r ;
  assign hwc_sum_g_d4 = cursor0_g | cursor1_g | cursor2_g |
                        cursor3_g | cursor4_g | cursor5_g ;
  assign hwc_sum_b_d4 = cursor0_b | cursor1_b | cursor2_b |
                        cursor3_b | cursor4_b | cursor5_b ;
  assign hwc_sum_t_d4 = cursor0_t & cursor1_t & cursor2_t &
                        cursor3_t & cursor4_t & cursor5_t ;

  reg     [3:0] sum_r_d5;
  reg     [3:0] sum_g_d5;
  reg     [3:0] sum_b_d5;

  always @(posedge bufged_aux)
  begin
    sum_b_d5 <= hwc_sum_t_d4 ? frame_b_d4 : hwc_sum_b_d4;
    sum_g_d5 <= hwc_sum_t_d4 ? frame_g_d4 : hwc_sum_g_d4;
    sum_r_d5 <= hwc_sum_t_d4 ? frame_r_d4 : hwc_sum_r_d4;
    VGA_R <= pixelblanking[5] ? 4'h0 : sum_r_d5;
    VGA_G <= pixelblanking[5] ? 4'h0 : sum_g_d5;
    VGA_B <= pixelblanking[5] ? 4'h0 : sum_b_d5;
  end

  //******************************************************************//
  // Frame buffer control and data ports 20h through 2Fh, all as r/w. //
  //******************************************************************//

  reg  [23:0] disp_start; // this is in words (pixels)
  reg  [11:0] disp_width; // this is in words (pixels)
  reg         disp_enable;
  reg         refr_enable;

  reg  [23:0] cpu_addr;
  reg   [7:0] cpu_data;
  reg         cpu_wr_stb;
  reg         crc_init_stb;

  always @(posedge bufged_div)
  begin
    if (write_strobe)
    begin
      if (port_id == 8'h20) disp_start[7:0] <= out_port;
      if (port_id == 8'h21) disp_start[15:8] <= out_port;
      if (port_id == 8'h22) disp_start[23:16] <= out_port;
      if (port_id == 8'h23) disp_width[7:0] <= out_port;
      if (port_id == 8'h24) disp_width[11:8] <= out_port[3:0];
      if (port_id == 8'h25) disp_enable <= out_port[0];
      if (port_id == 8'h26) disp_fade <= out_port[2:0];
      if (port_id == 8'h27) refr_enable <= out_port[0];
      if (port_id == 8'h28) cpu_addr[7:0] <= out_port;
      if (port_id == 8'h29) cpu_addr[15:8] <= out_port;
      if (port_id == 8'h2a) cpu_addr[23:16] <= out_port;
      if (port_id == 8'h2b) cpu_data <= out_port;
    end
    cpu_wr_stb <= write_strobe && (port_id == 8'h2c);
    crc_init_stb <= write_strobe && (port_id == 8'h2e);
    case (disp_fade)
      0 : fade_mult <= 4'b1000;
      1 : fade_mult <= 4'b0111;
      2 : fade_mult <= 4'b0110;
      3 : fade_mult <= 4'b0101;
      4 : fade_mult <= 4'b0100;
      5 : fade_mult <= 4'b0011;
      6 : fade_mult <= 4'b0001;
      7 : fade_mult <= 4'b0000;
    endcase
  end

  reg         cpu_write;
  reg         cpu_write_status;
  wire        clear_cpu_write;
  reg         ram_ready_status;
  wire        ram_ready;

  always @(negedge bufged_aux)
  begin
    if (cpu_wr_stb) cpu_write <= 1'b1;
    else if (clear_cpu_write) cpu_write <= 1'b0;
  end

  always @(posedge bufged_div) cpu_write_status <= cpu_write;
  always @(posedge bufged_div) ram_ready_status <= ram_ready;

  //******************************************************************//
  // Precompute some fetch parameters.  This is done here so the FSM  //
  // does not need to figure it out.  At best, I can get 1024 pixels  //
  // out of a row so a 1600 pixel line (max) will take take (at max)  //
  // three separate row accesses.  Lower resolution screens may take  //
  // fewer row accesses.                                              //
  //******************************************************************//

  reg  [23:0] fetch_offset;
  wire [12:0] fetch_rows;
  wire  [9:0] fetch_cols;

  reg  [11:0] disp_width_pipe; // pack into multiplier input
  reg  [11:0] vfetch_pipe; // pack into multiplier input
  always @(posedge bufged_aux) disp_width_pipe <= disp_width;
  always @(posedge bufged_aux) vfetch_pipe <= vfetch;
  always @(posedge bufged_aux) fetch_offset <= vfetch_pipe * disp_width_pipe;

  // This needed to be pipelined for performance
  // but I've preserved the original for reference.
  //
  // reg  [23:0] fetch_apoint;
  // always @(posedge bufged_aux) fetch_apoint <= fetch_offset + disp_start;

  reg  [12:0] helper2a;
  reg  [12:0] helper2b;
  wire [23:0] helper2;
  wire [23:0] fetch_apoint;

  always @(posedge bufged_aux) helper2a <= {1'b0,fetch_offset[11:0]} +
                                           {1'b0,disp_start[11:0]};
  always @(posedge bufged_aux) helper2b <= {1'b0,fetch_offset[23:12]} +
                                           {1'b0,disp_start[23:12]} +
                                           {12'h000,helper2a[12]};
  assign helper2 = {helper2b[11:0],helper2a[11:0]};
  assign fetch_apoint = helper2;

  assign fetch_cols = fetch_apoint[9:0];
  assign fetch_rows = fetch_apoint[22:10];

  reg  [12:0] fetch0_row;
  reg   [9:0] fetch0_col;
  reg  [10:0] fetch0_len;
  reg         fetch0_req;

  reg  [12:0] fetch1_row;
  reg   [9:0] fetch1_col;
  reg  [10:0] fetch1_len;
  reg         fetch1_req;

  reg  [12:0] fetch2_row;
  reg   [9:0] fetch2_col;
  reg  [10:0] fetch2_len;
  reg         fetch2_req;

  reg   [6:0] active_delay;

  // This needed to be pipelined for performance
  // but I've preserved the original for reference.
  //
  // reg  [11:0] helper0;
  // always @(posedge bufged_aux) helper0 <= fetch_cols + disp_width - 1;

  reg   [6:0] helper0a;
  reg   [6:0] helper0b;
  wire [11:0] helper0;

  always @(posedge bufged_aux) helper0a <= {1'b0,fetch_cols[5:0]}   + {1'b0,disp_width[5:0]} - 7'b0000001;
  always @(posedge bufged_aux) helper0b <= {3'b000,fetch_cols[9:6]} + {1'b0,disp_width[11:6]} + {6'b000000,helper0a[6]};
  assign helper0 = {helper0b[5:0],helper0a[5:0]};

  wire  [1:0] helper1;
  assign helper1 = helper0[11:10];

  always @(posedge bufged_aux)
  begin
    active_delay <= {active_delay[5:0],active};
    // easy stuff...
    fetch0_row <= fetch_rows;
    fetch1_row <= fetch_rows + 13'd1;
    fetch2_row <= fetch_rows + 13'd2;
    // more easy stuff...
    fetch0_col <= fetch_cols;
    fetch1_col <= 10'd0; // next row, start at 0
    fetch2_col <= 10'd0; // next row, start at 0
    // more complex
    case (helper1)
      // if helper1 is 2'b00, it means one fetch
      2'b00: begin
               fetch0_req <= disp_enable;
               fetch0_len <= disp_width;
               fetch1_req <= 1'd0;
               fetch1_len <= 11'd0;
               fetch2_req <= 1'b0;
               fetch2_len <= 11'd0;
             end
      // if helper1 is 2'b01, it means two fetch
      2'b01: begin
               fetch0_req <= disp_enable;
               fetch0_len <= 11'd1024 - fetch_cols;
               fetch1_req <= disp_enable;
               fetch1_len <= helper0[9:0] + 11'd1;
               fetch2_req <= 1'b0;
               fetch2_len <= 11'd0;
             end
      // if helper1 is 2'b1x, it means three fetch
      default: begin
               fetch0_req <= disp_enable;
               fetch0_len <= 11'd1024 - fetch_cols;
               fetch1_req <= disp_enable;
               fetch1_len <= 11'd1024;
               fetch2_req <= disp_enable;
               fetch2_len <= helper0[9:0] + 11'd1;
             end
    endcase
  end

  //******************************************************************//
  // Resync the precomputed information to the negative edge of the   //
  // clock so it's on the same clock as the FSM.  If I were to do     //
  // this design all over again, I would have changed the processor   //
  // clock to negative edge triggered to avoid the half-cycle timing  //
  // path that is created.                                            //
  //******************************************************************//

  reg  [12:0] nfetch0_row;
  reg   [9:0] nfetch0_col;
  reg  [10:0] nfetch0_len;
  reg         nfetch0_req;

  reg  [12:0] nfetch1_row;
  reg   [9:0] nfetch1_col;
  reg  [10:0] nfetch1_len;
  reg         nfetch1_req;

  reg  [12:0] nfetch2_row;
  reg   [9:0] nfetch2_col;
  reg  [10:0] nfetch2_len;
  reg         nfetch2_req;

  reg         nactive;
  reg  [23:0] ncpu_addr;

  // retime to negative edge for
  // the memory controller fsm

  always @(negedge bufged_aux)
  begin
    nfetch0_row <= fetch0_row;
    nfetch0_col <= fetch0_col;
    nfetch0_len <= fetch0_len;
    nfetch0_req <= fetch0_req;
    nfetch1_row <= fetch1_row;
    nfetch1_col <= fetch1_col;
    nfetch1_len <= fetch1_len;
    nfetch1_req <= fetch1_req;
    nfetch2_row <= fetch2_row;
    nfetch2_col <= fetch2_col;
    nfetch2_len <= fetch2_len;
    nfetch2_req <= fetch2_req;
    nactive <= active_delay[6];
    ncpu_addr <= cpu_addr;
  end

  //******************************************************************//
  // Implement parallel flash interface.                              //
  // Port 18h is the lo byte of the output address port.              //
  // Port 19h is the mi byte of the output address port.              //
  // Port 1Ah is the hi byte of the output address port.              //
  // Port 1Bh is the input data port.                                 //
  //******************************************************************//

  reg    [23:0] pfi_oaddr;
  wire          pfi_oaddrl_wr;
  wire          pfi_oaddrm_wr;
  wire          pfi_oaddrh_wr;

  assign pfi_oaddrl_wr = write_strobe & (port_id == 8'h18);
  assign pfi_oaddrm_wr = write_strobe & (port_id == 8'h19);
  assign pfi_oaddrh_wr = write_strobe & (port_id == 8'h1A);

  always @(posedge bufged_div)
  begin
    begin
      if (pfi_oaddrl_wr) pfi_oaddr[7:0] <= out_port[7:0];
      if (pfi_oaddrm_wr) pfi_oaddr[15:8] <= out_port[7:0];
      if (pfi_oaddrh_wr) pfi_oaddr[23:16] <= out_port[7:0];
    end
  end

  assign NF_A = pfi_oaddr[21:0];

  assign FPGA_INIT_B = 1'b0; // disable
  assign SPI_SS_B = 1'b1; // disable
  assign ALT_SS_B = 1'b1; // disable
  assign NF_BYTE = 1'b0; // byte mode
  assign NF_RP = 1'b1; // disable
  assign NF_WP = 1'b1; // disable
  assign NF_WE = 1'b1; // disable
  assign NF_CE = 1'b0; // enable
  assign NF_OE = 1'b0; // enable

  //******************************************************************//
  // Instantiate memory controller.  This has been modified from the  //
  // default design provided from MIG for the Spartan-3A Starter Kit. //
  //******************************************************************//

  wire   [31:0] mig_input_data;
  wire    [3:0] mig_input_mask;
  wire   [25:0] mig_input_addr;
  wire    [3:0] mig_input_cmds;
  wire          mig_input_burst_done;
  wire   [31:0] mig_output_data;
  wire          mig_output_data_valid;
  wire          mig_output_cmds_ack;
  wire          mig_output_init_val;
  wire          mig_output_ar_request;
  wire          mig_output_ar_done;
  wire          mig_output_wait_200us;
  wire          mig_output_rst;
  wire          mig_output_rst90;
  wire          mig_output_rst180;

  vlog_xst_bl4 vlog_xst_bl4_inst (
    .cntrl0_DDR2_DQ(SD_DQ),
    .cntrl0_DDR2_A(SD_A),
    .cntrl0_DDR2_BA(SD_BA),
    .cntrl0_DDR2_CK(SD_CK_P),
    .cntrl0_DDR2_CK_N(SD_CK_N),
    .cntrl0_DDR2_CKE(SD_CKE),
    .cntrl0_DDR2_CS_N(SD_CS),
    .cntrl0_DDR2_RAS_N(SD_RAS),
    .cntrl0_DDR2_CAS_N(SD_CAS),
    .cntrl0_DDR2_WE_N(SD_WE),
    .cntrl0_DDR2_ODT(SD_ODT),
    .cntrl0_DDR2_DM({SD_UDM,SD_LDM}),
    .cntrl0_DDR2_DQS_N({SD_UDQS_N,SD_LDQS_N}),
    .cntrl0_DDR2_DQS({SD_UDQS_P,SD_LDQS_P}),
    .cntrl0_rst_dqs_div_in(SD_LOOP_IN),
    .cntrl0_rst_dqs_div_out(SD_LOOP_OUT),

    .clock_out(bufged_aux),
    .clk90_out(bufged_a90),
    .ckdiv_out(bufged_div),
    .clocklock(clocklock),
    .reset_in_n(prelocked),
    .sys_clk_in(nubpre4x),
    .refr_enable(refr_enable),

    .mig_input_data(mig_input_data), // 32-bit input
    .mig_input_mask(mig_input_mask), // 4-bit input
    .mig_input_addr(mig_input_addr), // 26-bit input
    .mig_input_cmds(mig_input_cmds), // 4-bit input
    .mig_input_burst_done(mig_input_burst_done), // 1-bit input
    .mig_output_data(mig_output_data), // 32-bit output
    .mig_output_data_valid(mig_output_data_valid), // 1-bit output
    .mig_output_cmds_ack(mig_output_cmds_ack), // 1-bit output
    .mig_output_init_val(mig_output_init_val), // 1-bit output
    .mig_output_ar_request(mig_output_ar_request), // 1-bit output
    .mig_output_ar_done(mig_output_ar_done), // 1-bit output
    .mig_output_wait_200us(mig_output_wait_200us), // 1-bit output
    .mig_output_rst(mig_output_rst), // 1-bit output
    .mig_output_rst90(mig_output_rst90), // 1-bit output
    .mig_output_rst180(mig_output_rst180) // 1-bit output
  );

  //******************************************************************//
  // I'm sure I'll look back at this and laugh.  This uses an FSM to  //
  // control the sequences for initialization, processor data writes, //
  // and video data reads.  I am sure there are more elegant ways to  //
  // implement this function.  The best thing I can say about it is   //
  // that it works.  Note -- processor access during active display   //
  // refresh may cause some transient display noise.  The use model   //
  // here is for the processor to do a bulk copy of data before the   //
  // display fetch is enabled and then leave the buffer alone (with   //
  // the possible exeption of the vertical blanking interval).        //
  //******************************************************************//

  wire       master_reset;

  synchro #(.INITIALIZE("LOGIC1"))
  synchro_wait (.async(mig_output_rst180),.sync(master_reset),.clk(~bufged_aux));

  reg           refresh_active;
  wire          refresh_lockout;

  always @(negedge bufged_aux)
  begin
    if (master_reset) refresh_active <= 1'b0;
    else if (mig_output_ar_request) refresh_active <= 1'b1;
    else if (mig_output_ar_done) refresh_active <= 1'b0;
  end

  assign refresh_lockout = refresh_active | mig_output_ar_request;

  parameter [4:0] RESET = 5'd0;
  parameter [4:0] INIT0 = 5'd1;
  parameter [4:0] INIT1 = 5'd2;

  parameter [4:0] IDLED = 5'd3;

  // CPU Writes
  parameter [4:0] CPUW0 = 5'd4;
  parameter [4:0] CPUW1 = 5'd5;
  parameter [4:0] CPUW2 = 5'd6;
  parameter [4:0] CPUW3 = 5'd7;
  parameter [4:0] CPUW4 = 5'd8;
  parameter [4:0] CPUW5 = 5'd9;
  parameter [4:0] CPUW6 = 5'd10;
  parameter [4:0] CPUW7 = 5'd11;
  parameter [4:0] CPUW8 = 5'd12;
  parameter [4:0] CPUW9 = 5'd13;

  // Frame Buffer A Read (1st row of possible 3)
  parameter [4:0] FBAR0 = 5'd14;
  parameter [4:0] FBAR1 = 5'd15;
  parameter [4:0] FBAR2 = 5'd16;
  parameter [4:0] FBAR3 = 5'd17;
  parameter [4:0] FBAR4 = 5'd18;
  parameter [4:0] FBAR5 = 5'd19;

  // Frame Buffer B Read (2nd row of possible 3)
  parameter [4:0] FBBR0 = 5'd20;
  parameter [4:0] FBBR1 = 5'd21;
  parameter [4:0] FBBR2 = 5'd22;
  parameter [4:0] FBBR3 = 5'd23;
  parameter [4:0] FBBR4 = 5'd24;
  parameter [4:0] FBBR5 = 5'd25;

  // Frame Buffer C Read (3rd row of possible 3)
  parameter [4:0] FBCR0 = 5'd26;
  parameter [4:0] FBCR1 = 5'd27;
  parameter [4:0] FBCR2 = 5'd28;
  parameter [4:0] FBCR3 = 5'd29;
  parameter [4:0] FBCR4 = 5'd30;
  parameter [4:0] FBCR5 = 5'd31;

  reg  [4:0] state;
  reg [31:0] reg_input_data;
  reg  [3:0] reg_input_mask;
  reg [23:0] reg_input_addr;
  reg  [3:0] reg_input_cmds;
  reg        reg_input_burst_done;
  reg        reg_clr_cpuwr_strobe;
  reg  [8:0] reg_counter;
  wire       count_done;

  always @(negedge bufged_aux)
  begin
    reg_input_data <= {4{cpu_data}}; // spam it...
    if (master_reset)
    begin
      state <= RESET;
      reg_input_mask <= 4'h0;
      reg_input_addr <= 24'h0;
      reg_input_cmds <= 4'h0;
      reg_input_burst_done <= 1'b0;
      reg_clr_cpuwr_strobe <= 1'b1;
      reg_counter <= 9'b0;
    end
    else
    begin
      case (state)

        RESET: begin
                 state <= INIT0;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= 24'h0;
                 reg_input_cmds <= 4'h0;
                 reg_input_burst_done <= 1'b0;
                 reg_clr_cpuwr_strobe <= 1'b1;
                 reg_counter <= 9'b0;
               end

        INIT0: begin
                 state <= INIT1;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= 24'h0;
                 reg_input_cmds <= 4'h2;
                 reg_input_burst_done <= 1'b0;
                 reg_clr_cpuwr_strobe <= 1'b1;
                 reg_counter <= 9'b0;
               end

        INIT1: begin
                 state <= mig_output_init_val ? IDLED : INIT1;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= 24'h0;
                 reg_input_cmds <= 4'h0;
                 reg_input_burst_done <= 1'b0;
                 reg_clr_cpuwr_strobe <= 1'b1;
                 reg_counter <= 9'b0;
               end

        IDLED: begin
                 if (nactive && nfetch0_req) state <= FBAR0;
                 else if (refresh_lockout) state <= IDLED;
                 else state <= cpu_write ? CPUW0 : IDLED;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= 24'h0;
                 reg_input_cmds <= 4'h0;
                 reg_input_burst_done <= 1'b0;
                 reg_clr_cpuwr_strobe <= 1'b0;
                 reg_counter <= 9'b0;
               end

        CPUW0: begin
                 // nasty hack because there's no
                 // effective warning that refresh
                 // is about to occur... really wish
                 // there was no need to look at the
                 // refresh_lockout signal here!
                 state <= refresh_lockout ? IDLED : CPUW1;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= refresh_lockout ? 24'h0 : {ncpu_addr[23:11],1'b0,ncpu_addr[10:3],2'b00};
                 reg_input_cmds <= refresh_lockout ? 4'h0 : 4'h4;
                 reg_input_burst_done <= 1'b0;
                 reg_clr_cpuwr_strobe <= 1'b0;
                 reg_counter <= 9'b0;
               end

        CPUW1: begin
                 state <= CPUW2;
                 case (ncpu_addr[2:0])
                   3'b010 : reg_input_mask <= 4'b1110;
                   3'b011 : reg_input_mask <= 4'b1101;
                   3'b000 : reg_input_mask <= 4'b1011;
                   3'b001 : reg_input_mask <= 4'b0111;
                   default: reg_input_mask <= 4'b1111;
                 endcase
                 reg_input_addr <= {ncpu_addr[23:11],1'b0,ncpu_addr[10:3],2'b00};
                 reg_input_cmds <= 4'h4;
                 reg_input_burst_done <= 1'b0;
                 reg_clr_cpuwr_strobe <= 1'b0;
                 reg_counter <= 9'b0;
               end

        CPUW2: begin
                 state <= mig_output_cmds_ack ? CPUW3 : CPUW2;
                 case (ncpu_addr[2:0])
                   3'b110 : reg_input_mask <= 4'b1110;
                   3'b111 : reg_input_mask <= 4'b1101;
                   3'b100 : reg_input_mask <= 4'b1011;
                   3'b101 : reg_input_mask <= 4'b0111;
                   default: reg_input_mask <= 4'b1111;
                 endcase
                 reg_input_addr <= {ncpu_addr[23:11],1'b0,ncpu_addr[10:3],2'b00};
                 reg_input_cmds <= 4'h4;
                 reg_input_burst_done <= 1'b0;
                 reg_clr_cpuwr_strobe <= 1'b0;
                 reg_counter <= 9'b0;
               end

        CPUW3: begin
                 state <= CPUW4;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= {ncpu_addr[23:11],1'b0,ncpu_addr[10:3],2'b00};
                 reg_input_cmds <= 4'h4;
                 reg_input_burst_done <= 1'b0;
                 reg_clr_cpuwr_strobe <= 1'b0;
                 reg_counter <= 9'b0;
               end

        CPUW4: begin
                 state <= CPUW5;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= 24'h0;
                 reg_input_cmds <= 4'h4;
                 reg_input_burst_done <= 1'b1;
                 reg_clr_cpuwr_strobe <= 1'b0;
                 reg_counter <= 9'b0;
               end

        CPUW5: begin
                 state <= CPUW6;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= 24'h0;
                 reg_input_cmds <= 4'h4;
                 reg_input_burst_done <= 1'b1;
                 reg_clr_cpuwr_strobe <= 1'b0;
                 reg_counter <= 9'b0;
               end

        CPUW6: begin
                 state <= CPUW7;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= 24'h0;
                 reg_input_cmds <= 4'h4;
                 reg_input_burst_done <= 1'b0;
                 reg_clr_cpuwr_strobe <= 1'b1;
                 reg_counter <= 9'b0;
               end

        CPUW7: begin
                 state <= mig_output_cmds_ack ? CPUW7 : IDLED;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= 24'h0;
                 reg_input_cmds <= 4'h0;
                 reg_input_burst_done <= 1'b0;
                 reg_clr_cpuwr_strobe <= 1'b0;
                 reg_counter <= 9'b0;
               end

        FBAR0: begin
                 // nasty hack because there's no
                 // effective warning that refresh
                 // is about to occur... really wish
                 // there was no need to look at the
                 // refresh_lockout signal here!
                 state <= refresh_lockout ? FBAR0 : FBAR1;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= refresh_lockout ? 24'h0 : {nfetch0_row[11:0],1'b0,nfetch0_col[9:0]};
                 reg_input_cmds <= refresh_lockout ? 4'h0 : 4'h6;
                 reg_input_burst_done <= 1'b0;
                 reg_clr_cpuwr_strobe <= 1'b0;
                 reg_counter <= nfetch0_len[10:2];
               end

        FBAR1: begin
                 state <= mig_output_cmds_ack ? FBAR2 : FBAR1;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= {nfetch0_row[11:0],1'b0,nfetch0_col[9:0]};
                 reg_input_cmds <= 4'h6;
                 reg_input_burst_done <= 1'b0;
                 reg_clr_cpuwr_strobe <= 1'b0;
                 reg_counter <= nfetch0_len[10:2];
               end

        FBAR2: begin
                 state <= FBAR3;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= reg_input_addr;
                 reg_input_cmds <= 4'h6;
                 reg_input_burst_done <= 1'b0;
                 reg_clr_cpuwr_strobe <= 1'b0;
                 reg_counter <= reg_counter - 9'd1;
               end

        FBAR3: begin
                 state <= count_done ? FBAR4 : FBAR2;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= reg_input_addr + 4;
                 reg_input_cmds <= 4'h6;
                 reg_input_burst_done <= count_done;
                 reg_clr_cpuwr_strobe <= 1'b0;
                 reg_counter <= reg_counter;
               end

        FBAR4: begin
                 state <= FBAR5;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= 24'h0;
                 reg_input_cmds <= 4'h0;
                 reg_input_burst_done <= 1'b1;
                 reg_clr_cpuwr_strobe <= 1'b0;
                 reg_counter <= 9'b0;
               end

        FBAR5: begin
                 if (mig_output_cmds_ack) state <= FBAR5;
                 else if (nfetch1_req) state <= FBBR0;
                 else state <= IDLED;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= 24'h0;
                 reg_input_cmds <= 4'h0;
                 reg_input_burst_done <= 1'b0;
                 reg_clr_cpuwr_strobe <= 1'b0;
                 reg_counter <= 9'b0;
               end

        FBBR0: begin
                 // nasty hack because there's no
                 // effective warning that refresh
                 // is about to occur... really wish
                 // there was no need to look at the
                 // refresh_lockout signal here!
                 state <= refresh_lockout ? FBBR0 : FBBR1;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= refresh_lockout ? 24'h0 : {nfetch1_row[11:0],1'b0,nfetch1_col[9:0]};
                 reg_input_cmds <= refresh_lockout ? 4'h0 : 4'h6;
                 reg_input_burst_done <= 1'b0;
                 reg_clr_cpuwr_strobe <= 1'b0;
                 reg_counter <= nfetch1_len[10:2];
               end

        FBBR1: begin
                 state <= mig_output_cmds_ack ? FBBR2 : FBBR1;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= {nfetch1_row[11:0],1'b0,nfetch1_col[9:0]};
                 reg_input_cmds <= 4'h6;
                 reg_input_burst_done <= 1'b0;
                 reg_clr_cpuwr_strobe <= 1'b0;
                 reg_counter <= nfetch1_len[10:2];
               end

        FBBR2: begin
                 state <= FBBR3;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= reg_input_addr;
                 reg_input_cmds <= 4'h6;
                 reg_input_burst_done <= 1'b0;
                 reg_clr_cpuwr_strobe <= 1'b0;
                 reg_counter <= reg_counter - 9'd1;
               end

        FBBR3: begin
                 state <= count_done ? FBBR4 : FBBR2;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= reg_input_addr + 4;
                 reg_input_cmds <= 4'h6;
                 reg_input_burst_done <= count_done;
                 reg_clr_cpuwr_strobe <= 1'b0;
                 reg_counter <= reg_counter;
               end

        FBBR4: begin
                 state <= FBBR5;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= 24'h0;
                 reg_input_cmds <= 4'h0;
                 reg_input_burst_done <= 1'b1;
                 reg_clr_cpuwr_strobe <= 1'b0;
                 reg_counter <= 9'b0;
               end

        FBBR5: begin
                 if (mig_output_cmds_ack) state <= FBBR5;
                 else if (nfetch2_req) state <= FBCR0;
                 else state <= IDLED;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= 24'h0;
                 reg_input_cmds <= 4'h0;
                 reg_input_burst_done <= 1'b0;
                 reg_clr_cpuwr_strobe <= 1'b0;
                 reg_counter <= 9'b0;
               end

        FBCR0: begin
                 // nasty hack because there's no
                 // effective warning that refresh
                 // is about to occur... really wish
                 // there was no need to look at the
                 // refresh_lockout signal here!
                 state <= refresh_lockout ? FBCR0 : FBCR1;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= refresh_lockout ? 24'h0 : {nfetch2_row[11:0],1'b0,nfetch2_col[9:0]};
                 reg_input_cmds <= refresh_lockout ? 4'h0 : 4'h6;
                 reg_input_burst_done <= 1'b0;
                 reg_clr_cpuwr_strobe <= 1'b0;
                 reg_counter <= nfetch2_len[10:2];
               end

        FBCR1: begin
                 state <= mig_output_cmds_ack ? FBCR2 : FBCR1;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= {nfetch2_row[11:0],1'b0,nfetch2_col[9:0]};
                 reg_input_cmds <= 4'h6;
                 reg_input_burst_done <= 1'b0;
                 reg_clr_cpuwr_strobe <= 1'b0;
                 reg_counter <= nfetch2_len[10:2];
               end

        FBCR2: begin
                 state <= FBCR3;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= reg_input_addr;
                 reg_input_cmds <= 4'h6;
                 reg_input_burst_done <= 1'b0;
                 reg_clr_cpuwr_strobe <= 1'b0;
                 reg_counter <= reg_counter - 9'd1;
               end

        FBCR3: begin
                 state <= count_done ? FBCR4 : FBCR2;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= reg_input_addr + 4;
                 reg_input_cmds <= 4'h6;
                 reg_input_burst_done <= count_done;
                 reg_clr_cpuwr_strobe <= 1'b0;
                 reg_counter <= reg_counter;
               end

        FBCR4: begin
                 state <= FBCR5;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= 24'h0;
                 reg_input_cmds <= 4'h0;
                 reg_input_burst_done <= 1'b1;
                 reg_clr_cpuwr_strobe <= 1'b0;
                 reg_counter <= 9'b0;
               end

        FBCR5: begin
                 if (mig_output_cmds_ack) state <= FBCR5;
                 else state <= IDLED;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= 24'h0;
                 reg_input_cmds <= 4'h0;
                 reg_input_burst_done <= 1'b0;
                 reg_clr_cpuwr_strobe <= 1'b0;
                 reg_counter <= 9'b0;
               end

        default: begin
                 state <= RESET;
                 reg_input_mask <= 4'h0;
                 reg_input_addr <= 24'h0;
                 reg_input_cmds <= 4'h0;
                 reg_input_burst_done <= 1'b0;
                 reg_clr_cpuwr_strobe <= 1'b1;
                 reg_counter <= 9'b0;
               end

      endcase
    end
  end

  assign count_done = (reg_counter == 9'd0);

  assign clear_cpu_write = reg_clr_cpuwr_strobe;
  assign ram_ready = mig_output_init_val;

  assign mig_input_cmds = reg_input_cmds;
  assign mig_input_addr = {reg_input_addr,2'b00};
  assign mig_input_burst_done = reg_input_burst_done;

  reg [31:0] shift_input_data;
  reg  [3:0] shift_input_mask;

  always @(posedge bufged_a90) shift_input_data <= reg_input_data;
  always @(posedge bufged_a90) shift_input_mask <= reg_input_mask;

  assign mig_input_data = shift_input_data;
  assign mig_input_mask = shift_input_mask;

  //******************************************************************//
  // Implement write logic from external ram into the line buffers.   //
  //******************************************************************//

  reg    [31:0] dly_output_data;
  reg           dly_output_valid;

  always @(posedge bufged_a90) dly_output_data <= mig_output_data;
  always @(posedge bufged_a90) dly_output_valid <= mig_output_data_valid;

  reg    [31:0] output_data;
  reg           output_valid;

  always @(posedge bufged_aux) output_data <= dly_output_data;
  always @(posedge bufged_aux) output_valid <= dly_output_valid;

  reg     [9:0] write_ptr;

  always @(posedge bufged_aux)
  begin
    if (active_delay[5]) write_ptr <= 0;
    else if (output_valid) write_ptr <= write_ptr + 1;
  end

  assign memorydata = {output_data[15:0],output_data[31:16]};
  assign memoryaddr = write_ptr;
  assign memorywrite0 = output_valid && !vfetch[0];
  assign memorywrite1 = output_valid &&  vfetch[0];

  //******************************************************************//
  // CRC32 checking on each nibble as it exits video line buffers...  //
  //******************************************************************//

  wire   [31:0] crc32_red;
  reg    [31:0] crc32_red_cpu;

  crc32_4bit crc32_4bit_red (
    .crc_reg(crc32_red),
    .d(crc_r_d3),
    .init(crc_vsync_rst[3]),
    .d_valid(crc_dval),
    .clk(bufged_aux));
  always @(posedge bufged_div) crc32_red_cpu <= crc32_red;

  wire   [31:0] crc32_grn;
  reg    [31:0] crc32_grn_cpu;

  crc32_4bit crc32_4bit_grn (
    .crc_reg(crc32_grn),
    .d(crc_g_d3),
    .init(crc_vsync_rst[3]),
    .d_valid(crc_dval),
    .clk(bufged_aux));
  always @(posedge bufged_div) crc32_grn_cpu <= crc32_grn;

  wire   [31:0] crc32_blu;
  reg    [31:0] crc32_blu_cpu;

  crc32_4bit crc32_4bit_blu (
    .crc_reg(crc32_blu),
    .d(crc_b_d3),
    .init(crc_vsync_rst[3]),
    .d_valid(crc_dval),
    .clk(bufged_aux));
  always @(posedge bufged_div) crc32_blu_cpu <= crc32_blu;

  wire   [31:0] crc32_int;
  reg    [31:0] crc32_int_cpu;

  crc32_4bit crc32_4bit_int (
    .crc_reg(crc32_int),
    .d(crc_x_d3),
    .init(crc_vsync_rst[3]),
    .d_valid(crc_dval),
    .clk(bufged_aux));
  always @(posedge bufged_div) crc32_int_cpu <= crc32_int;

  //******************************************************************//
  // CRC32 checking on each nibble as it enters frame buffer...       //
  //******************************************************************//

  wire   [31:0] crc32_rfb;

  crc32_4bit crc32_4bit_rfb (
    .crc_reg(crc32_rfb),
    .d(cpu_data[3:0]),
    .init(crc_init_stb),
    .d_valid((cpu_wr_stb && !cpu_addr[0])),
    .clk(bufged_div));

  wire   [31:0] crc32_gfb;

  crc32_4bit crc32_4bit_gfb (
    .crc_reg(crc32_gfb),
    .d(cpu_data[7:4]),
    .init(crc_init_stb),
    .d_valid((cpu_wr_stb && !cpu_addr[0])),
    .clk(bufged_div));

  wire   [31:0] crc32_bfb;

  crc32_4bit crc32_4bit_bfb (
    .crc_reg(crc32_bfb),
    .d(cpu_data[3:0]),
    .init(crc_init_stb),
    .d_valid((cpu_wr_stb && cpu_addr[0])),
    .clk(bufged_div));

  wire   [31:0] crc32_ifb;

  crc32_4bit crc32_4bit_ifb (
    .crc_reg(crc32_ifb),
    .d(cpu_data[7:4]),
    .init(crc_init_stb),
    .d_valid((cpu_wr_stb && cpu_addr[0])),
    .clk(bufged_div));

  //******************************************************************//
  // Implement registers to control two hardware cursors.  These are  //
  // here for eye-candy effect to attract attention and at this time  //
  // are not pulling their data from the main memory.  Instead, the   //
  // image data is pre-initialized in block memory.  Cursor 0 and 1   //
  // share the same image data (block ram).  The data is output 2bpp  //
  // and then followed by a color look-up table.                      //
  //******************************************************************//

  reg    [11:0] cursor0_xpos = 0;
  reg    [11:0] cursor0_ypos = 0;
  reg    [12:0] cursor0_trgb0 = 13'b1_0000_0000_0000;
  reg    [12:0] cursor0_trgb1 = 13'b1_0000_0000_0000;
  reg    [12:0] cursor0_trgb2 = 13'b0_1001_0000_0011;
  reg    [12:0] cursor0_trgb3 = 13'b1_0000_0000_0000;
  reg    [11:0] cursor1_xpos = 800;
  reg    [11:0] cursor1_ypos = 600;

  always @(posedge bufged_div)
  begin
    if (write_strobe)
    begin
      if (port_id == 8'h80) cursor0_xpos[ 7: 0] <= out_port;
      if (port_id == 8'h81) cursor0_xpos[11: 8] <= out_port[3:0];
      if (port_id == 8'h82) cursor0_ypos[ 7: 0] <= out_port;
      if (port_id == 8'h83) cursor0_ypos[11: 8] <= out_port[3:0];
      if (port_id == 8'h84) cursor0_trgb0[ 7: 0] <= out_port;
      if (port_id == 8'h85) cursor0_trgb0[12: 8] <= out_port[4:0];
      if (port_id == 8'h86) cursor0_trgb1[ 7: 0] <= out_port;
      if (port_id == 8'h87) cursor0_trgb1[12: 8] <= out_port[4:0];
      if (port_id == 8'h88) cursor0_trgb2[ 7: 0] <= out_port;
      if (port_id == 8'h89) cursor0_trgb2[12: 8] <= out_port[4:0];
      if (port_id == 8'h8a) cursor0_trgb3[ 7: 0] <= out_port;
      if (port_id == 8'h8b) cursor0_trgb3[12: 8] <= out_port[4:0];
      if (port_id == 8'h8c) cursor1_xpos[ 7: 0] <= out_port;
      if (port_id == 8'h8d) cursor1_xpos[11: 8] <= out_port[3:0];
      if (port_id == 8'h8e) cursor1_ypos[ 7: 0] <= out_port;
      if (port_id == 8'h8f) cursor1_ypos[11: 8] <= out_port[3:0];
    end
  end

  reg    [11:0] hcount_hwc0_pipe;
  reg    [11:0] vcount_hwc0_pipe;

  reg    [11:0] cursor0_xidx;
  reg    [11:0] cursor0_yidx;
  wire    [1:0] cursor0_data;
  wire          cursor0_mask;

  reg    [11:0] cursor1_xidx;
  reg    [11:0] cursor1_yidx;
  wire    [1:0] cursor1_data;
  wire          cursor1_mask;

  always @(posedge bufged_aux) hcount_hwc0_pipe <= hcount;
  always @(posedge bufged_aux) vcount_hwc0_pipe <= vcount;

  always @(posedge bufged_aux) cursor0_xidx <= hcount_hwc0_pipe - cursor0_xpos;
  always @(posedge bufged_aux) cursor0_yidx <= vcount_hwc0_pipe - cursor0_ypos;
  always @(posedge bufged_aux) cursor1_xidx <= hcount_hwc0_pipe - cursor1_xpos;
  always @(posedge bufged_aux) cursor1_yidx <= vcount_hwc0_pipe - cursor1_ypos;

  cursor_pair #(.IMAGE("XILINX"))
  cursor_pair_inst0 (
    .xidx0(cursor0_xidx),
    .yidx0(cursor0_yidx),
    .data0(cursor0_data),
    .mask0(cursor0_mask),
    .xidx1(cursor1_xidx),
    .yidx1(cursor1_yidx),
    .data1(cursor1_data),
    .mask1(cursor1_mask),
    .clk(bufged_aux));

  always @(posedge bufged_aux)
  begin
    if (cursor0_mask)
    begin
      cursor0_r <= 4'h0;
      cursor0_g <= 4'h0;
      cursor0_b <= 4'h0;
      cursor0_t <= 1'b1;
    end
    else
    begin
      case (cursor0_data)
      0: begin
           cursor0_r <= cursor0_trgb0[11: 8];
           cursor0_g <= cursor0_trgb0[ 7: 4];
           cursor0_b <= cursor0_trgb0[ 3: 0];
           cursor0_t <= cursor0_trgb0[12];
         end
      1: begin
           cursor0_r <= cursor0_trgb1[11: 8];
           cursor0_g <= cursor0_trgb1[ 7: 4];
           cursor0_b <= cursor0_trgb1[ 3: 0];
           cursor0_t <= cursor0_trgb1[12];
         end
      2: begin
           cursor0_r <= cursor0_trgb2[11: 8];
           cursor0_g <= cursor0_trgb2[ 7: 4];
           cursor0_b <= cursor0_trgb2[ 3: 0];
           cursor0_t <= cursor0_trgb2[12];
         end
      3: begin
           cursor0_r <= cursor0_trgb3[11: 8];
           cursor0_g <= cursor0_trgb3[ 7: 4];
           cursor0_b <= cursor0_trgb3[ 3: 0];
           cursor0_t <= cursor0_trgb3[12];
         end
      endcase
    end
  end

  always @(posedge bufged_aux)
  begin
    if (cursor1_mask)
    begin
      cursor1_r <= 4'h0;
      cursor1_g <= 4'h0;
      cursor1_b <= 4'h0;
      cursor1_t <= 1'b1;
    end
    else
    begin
      case (cursor1_data)
      0: begin
           cursor1_r <= cursor0_trgb0[11: 8];
           cursor1_g <= cursor0_trgb0[ 7: 4];
           cursor1_b <= cursor0_trgb0[ 3: 0];
           cursor1_t <= cursor0_trgb0[12];
         end
      1: begin
           cursor1_r <= cursor0_trgb1[11: 8];
           cursor1_g <= cursor0_trgb1[ 7: 4];
           cursor1_b <= cursor0_trgb1[ 3: 0];
           cursor1_t <= cursor0_trgb1[12];
         end
      2: begin
           cursor1_r <= cursor0_trgb2[11: 8];
           cursor1_g <= cursor0_trgb2[ 7: 4];
           cursor1_b <= cursor0_trgb2[ 3: 0];
           cursor1_t <= cursor0_trgb2[12];
         end
      3: begin
           cursor1_r <= cursor0_trgb3[11: 8];
           cursor1_g <= cursor0_trgb3[ 7: 4];
           cursor1_b <= cursor0_trgb3[ 3: 0];
           cursor1_t <= cursor0_trgb3[12];
         end
      endcase
    end
  end

  //******************************************************************//
  // Implement registers to control two hardware cursors.  These are  //
  // here for eye-candy effect to attract attention and at this time  //
  // are not pulling their data from the main memory.  Instead, the   //
  // image data is pre-initialized in block memory.  Cursor 2 and 3   //
  // share the same image data (block ram).  The data is output 2bpp  //
  // and then followed by a color look-up table.                      //
  //******************************************************************//

  reg    [11:0] cursor2_xpos = 800;
  reg    [11:0] cursor2_ypos = 0;
  reg    [12:0] cursor2_trgb0 = 13'b0_0000_0000_0000;
  reg    [12:0] cursor2_trgb1 = 13'b0_1111_1111_1111;
  reg    [12:0] cursor2_trgb2 = 13'b1_0000_0000_0000;
  reg    [12:0] cursor2_trgb3 = 13'b0_1111_1101_0010;
  reg    [11:0] cursor3_xpos = 0;
  reg    [11:0] cursor3_ypos = 600;

  always @(posedge bufged_div)
  begin
    if (write_strobe)
    begin
      if (port_id == 8'h90) cursor2_xpos[ 7: 0] <= out_port;
      if (port_id == 8'h91) cursor2_xpos[11: 8] <= out_port[3:0];
      if (port_id == 8'h92) cursor2_ypos[ 7: 0] <= out_port;
      if (port_id == 8'h93) cursor2_ypos[11: 8] <= out_port[3:0];
      if (port_id == 8'h94) cursor2_trgb0[ 7: 0] <= out_port;
      if (port_id == 8'h95) cursor2_trgb0[12: 8] <= out_port[4:0];
      if (port_id == 8'h96) cursor2_trgb1[ 7: 0] <= out_port;
      if (port_id == 8'h97) cursor2_trgb1[12: 8] <= out_port[4:0];
      if (port_id == 8'h98) cursor2_trgb2[ 7: 0] <= out_port;
      if (port_id == 8'h99) cursor2_trgb2[12: 8] <= out_port[4:0];
      if (port_id == 8'h9a) cursor2_trgb3[ 7: 0] <= out_port;
      if (port_id == 8'h9b) cursor2_trgb3[12: 8] <= out_port[4:0];
      if (port_id == 8'h9c) cursor3_xpos[ 7: 0] <= out_port;
      if (port_id == 8'h9d) cursor3_xpos[11: 8] <= out_port[3:0];
      if (port_id == 8'h9e) cursor3_ypos[ 7: 0] <= out_port;
      if (port_id == 8'h9f) cursor3_ypos[11: 8] <= out_port[3:0];
    end
  end

  reg    [11:0] hcount_hwc2_pipe;
  reg    [11:0] vcount_hwc2_pipe;

  reg    [11:0] cursor2_xidx;
  reg    [11:0] cursor2_yidx;
  wire    [1:0] cursor2_data;
  wire          cursor2_mask;

  reg    [11:0] cursor3_xidx;
  reg    [11:0] cursor3_yidx;
  wire    [1:0] cursor3_data;
  wire          cursor3_mask;

  always @(posedge bufged_aux) hcount_hwc2_pipe <= hcount;
  always @(posedge bufged_aux) vcount_hwc2_pipe <= vcount;

  always @(posedge bufged_aux) cursor2_xidx <= hcount_hwc2_pipe - cursor2_xpos;
  always @(posedge bufged_aux) cursor2_yidx <= vcount_hwc2_pipe - cursor2_ypos;
  always @(posedge bufged_aux) cursor3_xidx <= hcount_hwc2_pipe - cursor3_xpos;
  always @(posedge bufged_aux) cursor3_yidx <= vcount_hwc2_pipe - cursor3_ypos;

  cursor_pair #(.IMAGE("SPARTAN3A"))
  cursor_pair_inst1 (
    .xidx0(cursor2_xidx),
    .yidx0(cursor2_yidx),
    .data0(cursor2_data),
    .mask0(cursor2_mask),
    .xidx1(cursor3_xidx),
    .yidx1(cursor3_yidx),
    .data1(cursor3_data),
    .mask1(cursor3_mask),
    .clk(bufged_aux));

  always @(posedge bufged_aux)
  begin
    if (cursor2_mask)
    begin
      cursor2_r <= 4'h0;
      cursor2_g <= 4'h0;
      cursor2_b <= 4'h0;
      cursor2_t <= 1'b1;
    end
    else
    begin
      case (cursor2_data)
      0: begin
           cursor2_r <= cursor2_trgb0[11: 8];
           cursor2_g <= cursor2_trgb0[ 7: 4];
           cursor2_b <= cursor2_trgb0[ 3: 0];
           cursor2_t <= cursor2_trgb0[12];
         end
      1: begin
           cursor2_r <= cursor2_trgb1[11: 8];
           cursor2_g <= cursor2_trgb1[ 7: 4];
           cursor2_b <= cursor2_trgb1[ 3: 0];
           cursor2_t <= cursor2_trgb1[12];
         end
      2: begin
           cursor2_r <= cursor2_trgb2[11: 8];
           cursor2_g <= cursor2_trgb2[ 7: 4];
           cursor2_b <= cursor2_trgb2[ 3: 0];
           cursor2_t <= cursor2_trgb2[12];
         end
      3: begin
           cursor2_r <= cursor2_trgb3[11: 8];
           cursor2_g <= cursor2_trgb3[ 7: 4];
           cursor2_b <= cursor2_trgb3[ 3: 0];
           cursor2_t <= cursor2_trgb3[12];
         end
      endcase
    end
  end

  always @(posedge bufged_aux)
  begin
    if (cursor3_mask)
    begin
      cursor3_r <= 4'h0;
      cursor3_g <= 4'h0;
      cursor3_b <= 4'h0;
      cursor3_t <= 1'b1;
    end
    else
    begin
      case (cursor3_data)
      0: begin
           cursor3_r <= cursor2_trgb0[11: 8];
           cursor3_g <= cursor2_trgb0[ 7: 4];
           cursor3_b <= cursor2_trgb0[ 3: 0];
           cursor3_t <= cursor2_trgb0[12];
         end
      1: begin
           cursor3_r <= cursor2_trgb1[11: 8];
           cursor3_g <= cursor2_trgb1[ 7: 4];
           cursor3_b <= cursor2_trgb1[ 3: 0];
           cursor3_t <= cursor2_trgb1[12];
         end
      2: begin
           cursor3_r <= cursor2_trgb2[11: 8];
           cursor3_g <= cursor2_trgb2[ 7: 4];
           cursor3_b <= cursor2_trgb2[ 3: 0];
           cursor3_t <= cursor2_trgb2[12];
         end
      3: begin
           cursor3_r <= cursor2_trgb3[11: 8];
           cursor3_g <= cursor2_trgb3[ 7: 4];
           cursor3_b <= cursor2_trgb3[ 3: 0];
           cursor3_t <= cursor2_trgb3[12];
         end
      endcase
    end
  end

  //******************************************************************//
  // Implement registers to control two hardware cursors.  These are  //
  // here for eye-candy effect to attract attention and at this time  //
  // are not pulling their data from the main memory.  Instead, the   //
  // image data is pre-initialized in block memory.  Cursor 4 and 5   //
  // share the same image data (block ram).  The data is output 2bpp  //
  // and then followed by a color look-up table.                      //
  //******************************************************************//

  reg    [11:0] cursor4_xpos = 400;
  reg    [11:0] cursor4_ypos = 300;
  reg    [12:0] cursor4_trgb0 = 13'b0_0000_0000_0000;
  reg    [12:0] cursor4_trgb1 = 13'b1_0000_0000_0000;
  reg    [12:0] cursor4_trgb2 = 13'b0_1001_0000_0011;
  reg    [12:0] cursor4_trgb3 = 13'b0_1000_1000_1000;
  reg    [11:0] cursor5_xpos = 1200;
  reg    [11:0] cursor5_ypos = 900;

  always @(posedge bufged_div)
  begin
    if (write_strobe)
    begin
      if (port_id == 8'ha0) cursor4_xpos[ 7: 0] <= out_port;
      if (port_id == 8'ha1) cursor4_xpos[11: 8] <= out_port[3:0];
      if (port_id == 8'ha2) cursor4_ypos[ 7: 0] <= out_port;
      if (port_id == 8'ha3) cursor4_ypos[11: 8] <= out_port[3:0];
      if (port_id == 8'ha4) cursor4_trgb0[ 7: 0] <= out_port;
      if (port_id == 8'ha5) cursor4_trgb0[12: 8] <= out_port[4:0];
      if (port_id == 8'ha6) cursor4_trgb1[ 7: 0] <= out_port;
      if (port_id == 8'ha7) cursor4_trgb1[12: 8] <= out_port[4:0];
      if (port_id == 8'ha8) cursor4_trgb2[ 7: 0] <= out_port;
      if (port_id == 8'ha9) cursor4_trgb2[12: 8] <= out_port[4:0];
      if (port_id == 8'haa) cursor4_trgb3[ 7: 0] <= out_port;
      if (port_id == 8'hab) cursor4_trgb3[12: 8] <= out_port[4:0];
      if (port_id == 8'hac) cursor5_xpos[ 7: 0] <= out_port;
      if (port_id == 8'had) cursor5_xpos[11: 8] <= out_port[3:0];
      if (port_id == 8'hae) cursor5_ypos[ 7: 0] <= out_port;
      if (port_id == 8'haf) cursor5_ypos[11: 8] <= out_port[3:0];
    end
  end

  reg    [11:0] hcount_hwc4_pipe;
  reg    [11:0] vcount_hwc4_pipe;

  reg    [11:0] cursor4_xidx;
  reg    [11:0] cursor4_yidx;
  wire    [1:0] cursor4_data;
  wire          cursor4_mask;

  reg    [11:0] cursor5_xidx;
  reg    [11:0] cursor5_yidx;
  wire    [1:0] cursor5_data;
  wire          cursor5_mask;

  always @(posedge bufged_aux) hcount_hwc4_pipe <= hcount;
  always @(posedge bufged_aux) vcount_hwc4_pipe <= vcount;

  always @(posedge bufged_aux) cursor4_xidx <= hcount_hwc4_pipe - cursor4_xpos;
  always @(posedge bufged_aux) cursor4_yidx <= vcount_hwc4_pipe - cursor4_ypos;
  always @(posedge bufged_aux) cursor5_xidx <= hcount_hwc4_pipe - cursor5_xpos;
  always @(posedge bufged_aux) cursor5_yidx <= vcount_hwc4_pipe - cursor5_ypos;

  cursor_pair #(.IMAGE("VIRTEX5"))
  cursor_pair_inst2 (
    .xidx0(cursor4_xidx),
    .yidx0(cursor4_yidx),
    .data0(cursor4_data),
    .mask0(cursor4_mask),
    .xidx1(cursor5_xidx),
    .yidx1(cursor5_yidx),
    .data1(cursor5_data),
    .mask1(cursor5_mask),
    .clk(bufged_aux));

  always @(posedge bufged_aux)
  begin
    if (cursor4_mask)
    begin
      cursor4_r <= 4'h0;
      cursor4_g <= 4'h0;
      cursor4_b <= 4'h0;
      cursor4_t <= 1'b1;
    end
    else
    begin
      case (cursor4_data)
      0: begin
           cursor4_r <= cursor4_trgb0[11: 8];
           cursor4_g <= cursor4_trgb0[ 7: 4];
           cursor4_b <= cursor4_trgb0[ 3: 0];
           cursor4_t <= cursor4_trgb0[12];
         end
      1: begin
           cursor4_r <= cursor4_trgb1[11: 8];
           cursor4_g <= cursor4_trgb1[ 7: 4];
           cursor4_b <= cursor4_trgb1[ 3: 0];
           cursor4_t <= cursor4_trgb1[12];
         end
      2: begin
           cursor4_r <= cursor4_trgb2[11: 8];
           cursor4_g <= cursor4_trgb2[ 7: 4];
           cursor4_b <= cursor4_trgb2[ 3: 0];
           cursor4_t <= cursor4_trgb2[12];
         end
      3: begin
           cursor4_r <= cursor4_trgb3[11: 8];
           cursor4_g <= cursor4_trgb3[ 7: 4];
           cursor4_b <= cursor4_trgb3[ 3: 0];
           cursor4_t <= cursor4_trgb3[12];
         end
      endcase
    end
  end

  always @(posedge bufged_aux)
  begin
    if (cursor5_mask)
    begin
      cursor5_r <= 4'h0;
      cursor5_g <= 4'h0;
      cursor5_b <= 4'h0;
      cursor5_t <= 1'b1;
    end
    else
    begin
      case (cursor5_data)
      0: begin
           cursor5_r <= cursor4_trgb0[11: 8];
           cursor5_g <= cursor4_trgb0[ 7: 4];
           cursor5_b <= cursor4_trgb0[ 3: 0];
           cursor5_t <= cursor4_trgb0[12];
         end
      1: begin
           cursor5_r <= cursor4_trgb1[11: 8];
           cursor5_g <= cursor4_trgb1[ 7: 4];
           cursor5_b <= cursor4_trgb1[ 3: 0];
           cursor5_t <= cursor4_trgb1[12];
         end
      2: begin
           cursor5_r <= cursor4_trgb2[11: 8];
           cursor5_g <= cursor4_trgb2[ 7: 4];
           cursor5_b <= cursor4_trgb2[ 3: 0];
           cursor5_t <= cursor4_trgb2[12];
         end
      3: begin
           cursor5_r <= cursor4_trgb3[11: 8];
           cursor5_g <= cursor4_trgb3[ 7: 4];
           cursor5_b <= cursor4_trgb3[ 3: 0];
           cursor5_t <= cursor4_trgb3[12];
         end
      endcase
    end
  end

  //******************************************************************//
  // Implement registers to interface with look-up table for demo.    //
  // F8h rd = autopilot address                                       //
  // F8h wr = autopilot address                                       //
  // F9h rd = autopilot data x-lo                                     //
  // F9h wr = not implemented                                         //
  // FAh rd = autopilot data x-hi                                     //
  // FAh wr = not implemented                                         //
  // FBh rd = autopilot data y-lo                                     //
  // FBh wr = not implemented                                         //
  // FCh rd = autopilot data y-hi                                     //
  // FCh wr = not implemented                                         //
  //******************************************************************//

  reg     [7:0] auto_addr; 
  wire   [11:0] auto_data_x; 
  wire   [11:0] auto_data_y; 

  always @(posedge bufged_div)
  begin
    if (write_strobe)
    begin
      if (port_id == 8'hf8) auto_addr <= out_port;
    end
  end

  autopilot autopilot_inst (
    .timestep(auto_addr),
    .xlocation(auto_data_x),
    .ylocation(auto_data_y),
    .clk(bufged_div));

  reg    [17:1] lfsr;

  initial lfsr <= 17'b1_1111_1111_1111_1111;
  always @ (posedge bufged_div) lfsr <= {lfsr[16:1], (lfsr[17] ^ lfsr[14])};

  //******************************************************************//
  // Implement input port logic.                                      //
  //******************************************************************//

  always @(posedge bufged_div)
  begin
    case (port_id)

      8'h00: in_port <= {4'b0000,sws_out};
      8'h01: in_port <= {4'b0000,btn_out};
      8'h02: in_port <= {4'b0000,rot_out};

      8'h03: in_port <= LCD_DB;
      8'h04: in_port <= {5'b00000,LCD_RS,LCD_RW,LCD_E};
      8'h05: in_port <= LED;

      8'h06: in_port <= 8'h00; // reserved
      8'h07: in_port <= {6'b000000,tc_hs_neg,tc_vs_neg};

      8'h08: in_port <= tc_hsblnk[7:0];
      8'h09: in_port <= {4'b0000,tc_hsblnk[11:8]};
      8'h0a: in_port <= tc_hssync[7:0];
      8'h0b: in_port <= {4'b0000,tc_hssync[11:8]};
      8'h0c: in_port <= tc_hesync[7:0];
      8'h0d: in_port <= {4'b0000,tc_hesync[11:8]};
      8'h0e: in_port <= tc_heblnk[7:0];
      8'h0f: in_port <= {4'b0000,tc_heblnk[11:8]};
      8'h10: in_port <= tc_vsblnk[7:0];
      8'h11: in_port <= {4'b0000,tc_vsblnk[11:8]};
      8'h12: in_port <= tc_vssync[7:0];
      8'h13: in_port <= {4'b0000,tc_vssync[11:8]};
      8'h14: in_port <= tc_vesync[7:0];
      8'h15: in_port <= {4'b0000,tc_vesync[11:8]};
      8'h16: in_port <= tc_veblnk[7:0];
      8'h17: in_port <= {4'b0000,tc_veblnk[11:8]};

      8'h18: in_port <= pfi_oaddr[7:0];
      8'h19: in_port <= pfi_oaddr[15:8];
      8'h1a: in_port <= pfi_oaddr[23:16];
      8'h1b: in_port <= NF_D;

      8'h20: in_port <= disp_start[7:0];
      8'h21: in_port <= disp_start[15:8];
      8'h22: in_port <= disp_start[23:16];
      8'h23: in_port <= disp_width[7:0];
      8'h24: in_port <= {4'b0000,disp_width[11:8]};
      8'h25: in_port <= {7'b0000000,disp_enable};
      8'h26: in_port <= {5'b00000,disp_fade};
      8'h27: in_port <= {7'b0000000,refr_enable};

      8'h28: in_port <= cpu_addr[7:0];
      8'h29: in_port <= cpu_addr[15:8];
      8'h2a: in_port <= cpu_addr[23:16];
      8'h2b: in_port <= cpu_data;
      8'h2c: in_port <= cpu_write_status;
      8'h2d: in_port <= ram_ready_status;
      8'h2e: in_port <= 8'h00; // reserved

      8'h60: in_port <= crc32_red_cpu[ 7: 0];
      8'h61: in_port <= crc32_red_cpu[15: 8];
      8'h62: in_port <= crc32_red_cpu[23:16];
      8'h63: in_port <= crc32_red_cpu[31:24];
      8'h64: in_port <= crc32_grn_cpu[ 7: 0];
      8'h65: in_port <= crc32_grn_cpu[15: 8];
      8'h66: in_port <= crc32_grn_cpu[23:16];
      8'h67: in_port <= crc32_grn_cpu[31:24];
      8'h68: in_port <= crc32_blu_cpu[ 7: 0];
      8'h69: in_port <= crc32_blu_cpu[15: 8];
      8'h6a: in_port <= crc32_blu_cpu[23:16];
      8'h6b: in_port <= crc32_blu_cpu[31:24];
      8'h6c: in_port <= crc32_int_cpu[ 7: 0];
      8'h6d: in_port <= crc32_int_cpu[15: 8];
      8'h6e: in_port <= crc32_int_cpu[23:16];
      8'h6f: in_port <= crc32_int_cpu[31:24];

      8'h70: in_port <= crc32_rfb[ 7: 0];
      8'h71: in_port <= crc32_rfb[15: 8];
      8'h72: in_port <= crc32_rfb[23:16];
      8'h73: in_port <= crc32_rfb[31:24];
      8'h74: in_port <= crc32_gfb[ 7: 0];
      8'h75: in_port <= crc32_gfb[15: 8];
      8'h76: in_port <= crc32_gfb[23:16];
      8'h77: in_port <= crc32_gfb[31:24];
      8'h78: in_port <= crc32_bfb[ 7: 0];
      8'h79: in_port <= crc32_bfb[15: 8];
      8'h7a: in_port <= crc32_bfb[23:16];
      8'h7b: in_port <= crc32_bfb[31:24];
      8'h7c: in_port <= crc32_ifb[ 7: 0];
      8'h7d: in_port <= crc32_ifb[15: 8];
      8'h7e: in_port <= crc32_ifb[23:16];
      8'h7f: in_port <= crc32_ifb[31:24];

      8'h80: in_port <= cursor0_xpos[ 7: 0];
      8'h81: in_port <= {4'b0000,cursor0_xpos[11: 8]};
      8'h82: in_port <= cursor0_ypos[ 7: 0];
      8'h83: in_port <= {4'b0000,cursor0_ypos[11: 8]};
      8'h84: in_port <= cursor0_trgb0[ 7: 0];
      8'h85: in_port <= {3'b000,cursor0_trgb0[12: 8]};
      8'h86: in_port <= cursor0_trgb1[ 7: 0];
      8'h87: in_port <= {3'b000,cursor0_trgb1[12: 8]};
      8'h88: in_port <= cursor0_trgb2[ 7: 0];
      8'h89: in_port <= {3'b000,cursor0_trgb2[12: 8]};
      8'h8a: in_port <= cursor0_trgb3[ 7: 0];
      8'h8b: in_port <= {3'b000,cursor0_trgb3[12: 8]};
      8'h8c: in_port <= cursor1_xpos[ 7: 0];
      8'h8d: in_port <= {4'b0000,cursor1_xpos[11: 8]};
      8'h8e: in_port <= cursor1_ypos[ 7: 0];
      8'h8f: in_port <= {4'b0000,cursor1_ypos[11: 8]};

      8'h90: in_port <= cursor2_xpos[ 7: 0];
      8'h91: in_port <= {4'b0000,cursor2_xpos[11: 8]};
      8'h92: in_port <= cursor2_ypos[ 7: 0];
      8'h93: in_port <= {4'b0000,cursor2_ypos[11: 8]};
      8'h94: in_port <= cursor2_trgb0[ 7: 0];
      8'h95: in_port <= {3'b000,cursor2_trgb0[12: 8]};
      8'h96: in_port <= cursor2_trgb1[ 7: 0];
      8'h97: in_port <= {3'b000,cursor2_trgb1[12: 8]};
      8'h98: in_port <= cursor2_trgb2[ 7: 0];
      8'h99: in_port <= {3'b000,cursor2_trgb2[12: 8]};
      8'h9a: in_port <= cursor2_trgb3[ 7: 0];
      8'h9b: in_port <= {3'b000,cursor2_trgb3[12: 8]};
      8'h9c: in_port <= cursor3_xpos[ 7: 0];
      8'h9d: in_port <= {4'b0000,cursor3_xpos[11: 8]};
      8'h9e: in_port <= cursor3_ypos[ 7: 0];
      8'h9f: in_port <= {4'b0000,cursor3_ypos[11: 8]};

      8'ha0: in_port <= cursor4_xpos[ 7: 0];
      8'ha1: in_port <= {4'b0000,cursor4_xpos[11: 8]};
      8'ha2: in_port <= cursor4_ypos[ 7: 0];
      8'ha3: in_port <= {4'b0000,cursor4_ypos[11: 8]};
      8'ha4: in_port <= cursor4_trgb0[ 7: 0];
      8'ha5: in_port <= {3'b000,cursor4_trgb0[12: 8]};
      8'ha6: in_port <= cursor4_trgb1[ 7: 0];
      8'ha7: in_port <= {3'b000,cursor4_trgb1[12: 8]};
      8'ha8: in_port <= cursor4_trgb2[ 7: 0];
      8'ha9: in_port <= {3'b000,cursor4_trgb2[12: 8]};
      8'haa: in_port <= cursor4_trgb3[ 7: 0];
      8'hab: in_port <= {3'b000,cursor4_trgb3[12: 8]};
      8'hac: in_port <= cursor5_xpos[ 7: 0];
      8'had: in_port <= {4'b0000,cursor5_xpos[11: 8]};
      8'hae: in_port <= cursor5_ypos[ 7: 0];
      8'haf: in_port <= {4'b0000,cursor5_ypos[11: 8]};

      8'hf0: in_port <= {7'b0000000,CRCERR};

      8'hf8: in_port <= auto_addr;
      8'hf9: in_port <= auto_data_x[7:0];
      8'hfa: in_port <= {4'b0000,auto_data_x[11:8]};
      8'hfb: in_port <= auto_data_y[7:0];
      8'hfc: in_port <= {4'b0000,auto_data_y[11:8]};
      8'hfd: in_port <= lfsr[8:1];

      8'hff: in_port <= 8'h00; // reserved
      default: in_port <= 8'h00;
    endcase
  end

  //******************************************************************//
  //                                                                  //
  //******************************************************************//

endmodule