//////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2007 Xilinx, Inc.
// This design is confidential and proprietary of Xilinx, All Rights Reserved.
//////////////////////////////////////////////////////////////////////////////
//   ____  ____
//  /   /\/   /
// /___/  \  /   Vendor:        Xilinx
// \   \   \/    Version:       1.0.0
//  \   \        Filename:      linebuf.v
//  /   /        Date Created:  July 1, 2007
// /___/   /\    Last Modified: July 1, 2007
// \   \  /  \
//  \___\/\___\
//
// Devices:   Spartan-3 Generation FPGA
// Purpose:   Width converting line buffer circuit
// Contact:   crabill@xilinx.com
// Reference: None
//
// Revision History:
//   Rev 1.0.0 - (crabill) First created July 1, 2007.
//
//////////////////////////////////////////////////////////////////////////////
//
// LIMITED WARRANTY AND DISCLAIMER. These designs are provided to you "as is".
// Xilinx and its licensors make and you receive no warranties or conditions,
// express, implied, statutory or otherwise, and Xilinx specifically disclaims
// any implied warranties of merchantability, non-infringement, or fitness for
// a particular purpose. Xilinx does not warrant that the functions contained
// in these designs will meet your requirements, or that the operation of
// these designs will be uninterrupted or error free, or that defects in the
// designs will be corrected. Furthermore, Xilinx does not warrant or make any
// representations regarding use or the results of the use of the designs in
// terms of correctness, accuracy, reliability, or otherwise.
//
// LIMITATION OF LIABILITY. In no event will Xilinx or its licensors be liable
// for any loss of data, lost profits, cost or procurement of substitute goods
// or services, or for any special, incidental, consequential, or indirect
// damages arising from the use or operation of the designs or accompanying
// documentation, however caused and on any theory of liability. This
// limitation will apply even if Xilinx has been advised of the possibility
// of such damage. This limitation shall apply not-withstanding the failure
// of the essential purpose of any limited remedies herein.
//
//////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2007 Xilinx, Inc.
// This design is confidential and proprietary of Xilinx, All Rights Reserved.
//////////////////////////////////////////////////////////////////////////////

`timescale 1 ns / 1 ps

module linebuf
  (
  input  wire [10:0] hcount,
  output reg  [15:0] pixeldata,
  input  wire        readenable,

  input  wire [31:0] memorydata,
  input  wire  [9:0] memoryaddr,
  input  wire        memorywrite,

  input  wire        clk
  );

  //******************************************************************//
  // Port A is organized as 2kx16 and Port B is organized as 1kx32.   //
  // Data is written into Port B from the frame buffer and read out   //
  // on Port A by the video timing logic.                             //
  //******************************************************************//
 
  wire [15:0] lba_out;

  RAMB16_S18_S36 LBA (
  .DOA(lba_out),
  .DOPA(),
  .ADDRA(hcount[9:0]),
  .CLKA(clk), 
  .DIA(16'h0000),
  .DIPA(2'b00),
  .ENA(readenable), 
  .WEA(1'b0),
  .SSRA(1'b0),
  .DOB(),
  .DOPB(), 
  .ADDRB(memoryaddr[8:0]), 
  .CLKB(clk), 
  .DIB(memorydata), 
  .DIPB(4'h0), 
  .ENB(!memoryaddr[9]), 
  .WEB(memorywrite),
  .SSRB(1'b0)
  );

  wire [15:0] lbb_out;

  RAMB16_S18_S36 LBB (
  .DOA(lbb_out),
  .DOPA(),
  .ADDRA(hcount[9:0]),
  .CLKA(clk), 
  .DIA(16'h0000),
  .DIPA(2'b00),
  .ENA(readenable), 
  .WEA(1'b0),
  .SSRA(1'b0),
  .DOB(),
  .DOPB(), 
  .ADDRB(memoryaddr[8:0]), 
  .CLKB(clk), 
  .DIB(memorydata), 
  .DIPB(4'h0), 
  .ENB(memoryaddr[9]), 
  .WEB(memorywrite),
  .SSRB(1'b0)
  );

  //******************************************************************//
  // Pipelined output mux for read data...                            //
  //******************************************************************//

  reg           select;

  always @(posedge clk)
  begin : omux
    select <= hcount[10];
    pixeldata <= select ? lbb_out : lba_out;
  end

  //******************************************************************//
  //                                                                  //
  //******************************************************************//

endmodule