///////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005 Xilinx, Inc.
// This design is confidential and proprietary of Xilinx, All Rights Reserved.
///////////////////////////////////////////////////////////////////////////////
//   ____  ____
//  /   /\/   /
// /___/  \  /   Vendor			: Xilinx
// \   \   \/    Version		: $Name: mig_v1_7_b7 $
//  \   \        Application		: MIG
//  /   /        Filename		: vlog_xst_bl4_cal_ctl.v
// /___/   /\    Date Last Modified	: $Date: 2007/02/10 10:57:50 $
// \   \  /  \   Date Created		: Mon May 2 2005
//  \___\/\___\
// Device	: Spartan-3/3A
// Design Name	: DDR2 SDRAM
// Purpose	: This module generates the select lines for the LUT delay 
//		  circuit that generate the required delay for the DQS with 
//		  respect to the DQ. It calculates the dealy of a LUT 
//		  dynalically by finding the number of LUTs in a clock phase. 
///////////////////////////////////////////////////////////////////////////////

`timescale 1ns/100ps

module vlog_xst_bl4_cal_ctl
  (
   input            clk,
   input            reset,
   input [31:0]     flop2,
   output reg [4:0] tapForDqs
   );


   localparam tap1        = 5'b01111;
   localparam tap2        = 5'b10111;
   localparam tap3        = 5'b11011;
   localparam tap4        = 5'b11101;
   localparam tap5        = 5'b11110;
   localparam tap6        = 5'b11111;
   localparam defaultTap  = 5'b11101;

   reg [5:0]        cnt; 
   reg [5:0]        cnt1;
   reg              trans_oneDtct;
   reg              trans_twoDtct;
   reg [4:0]        phase_cnt;
   reg [31:0]       tap_dly_reg;
   reg              enb_trans_two_dtct;
   reg [4:0]        tapForDqs1;
   reg		    reset_r;

   always @( posedge clk )
     reset_r <= reset;

   always @(posedge clk) begin
      if(reset_r)
        enb_trans_two_dtct <= 1'b0;
      else if(phase_cnt >= 5'd3)
        enb_trans_two_dtct <= 1'b1;
      else
        enb_trans_two_dtct <= 1'b0;
   end

   always @(posedge clk) begin
      if(reset_r)
        tap_dly_reg <= 32'd0;
      else if(cnt[5] == 1'b1)
        tap_dly_reg <= flop2;
      else
        tap_dly_reg <= tap_dly_reg;
   end

/************* Free Running Counter For Counting 32 States ********************/

   always @(posedge clk) begin
      if(reset_r || (cnt[5] == 1'b1))
        cnt[5:0] <= 6'b0;
      else
        cnt[5:0] <= cnt[5:0] + 1'b1;
   end

   always @(posedge clk) begin
      if(reset_r || (cnt1[5] == 1'b1))
        cnt1[5:0] <= 6'b0;
      else
        cnt1[5:0] <= cnt1[5:0] + 1'b1;
   end

   always @(posedge clk) begin
      if(reset_r || (cnt[5] == 1'b1)) begin
         phase_cnt <= 5'd0;
      end
      else if (trans_oneDtct & (~trans_twoDtct))
        phase_cnt <= phase_cnt + 1;
      else
        phase_cnt <= phase_cnt;
   end

/**************** Checking For The First Transition ***************************/
   always @(posedge clk) begin
      if(reset_r) begin
         trans_oneDtct <= 1'b0;
         trans_twoDtct <= 1'b0;
      end
      else if(cnt[5] == 1'b1) begin
         trans_oneDtct <= 1'b0;
         trans_twoDtct <= 1'b0;
      end
      else if (cnt[4:0] == 5'd0) begin
         if ((tap_dly_reg[0])) begin
            trans_oneDtct <= 1'b1;
            trans_twoDtct <= 1'b0;
         end
      end
      else if ((tap_dly_reg[cnt[4:0]]) & (trans_twoDtct == 1'b0)) begin
         if((trans_oneDtct == 1'b1) && (enb_trans_two_dtct) ) begin
            trans_twoDtct <= 1'b1;
         end
         else begin
            trans_oneDtct <= 1'b1;
         end
      end
   end
   always @(posedge clk) begin
      if(reset_r)
         tapForDqs1 <= defaultTap;
      else if(cnt1[4] && cnt1[3] &&
              cnt1[2] && cnt1[1] && cnt1[0]) begin
           if((trans_oneDtct == 1'b0) || (trans_twoDtct == 1'b0)
               || (phase_cnt > 5'd11))
              tapForDqs1 <= tap4;
           else if((phase_cnt > 5'd8))
              tapForDqs1 <= tap3;
           else
              tapForDqs1 <= tap2;
        end
      else
         tapForDqs1 <= tapForDqs1;
   end
  
  always @(posedge clk)
     tapForDqs  <= tapForDqs1;
endmodule
