///////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005 Xilinx, Inc.
// This design is confidential and proprietary of Xilinx, All Rights Reserved.
///////////////////////////////////////////////////////////////////////////////
//   ____  ____
//  /   /\/   /
// /___/  \  /   Vendor			: Xilinx
// \   \   \/    Version		: $Name: mig_v1_7_b7 $
//  \   \        Application		: MIG
//  /   /        Filename		: vlog_xst_bl4_cmd_fsm_0.v
// /___/   /\    Date Last Modified	: $Date: 2007/02/19 12:03:15 $
// \   \  /  \   Date Created		: Mon May 2 2005
//  \___\/\___\
// Device	: Spartan-3/3A
// Design Name	: DDR2 SDRAM
// Purpose	: This module generate the commands to the controller.
///////////////////////////////////////////////////////////////////////////////

`timescale 1ns/100ps
`include "../source/sdramtst/vlog_xst_bl4_parameters_0.v"
module vlog_xst_bl4_cmd_fsm_0
  (
   input             clk,
   input             clk90,
   input             rst180,
   input             rst90,
   input             cmd_ack,
   input             cnt_roll,
   input             refresh_done,
   input             init_val,
   output reg        r_w,
   output reg        addr_inc,
   output            addr_rst,
   output reg [3:0]  u_cmd,
   output            lfsr_rst,
   input             auto_ref_req
   );


   parameter [4:0]   rst_state      = 0,
                     init_start     = 1,
                     init           = 2,
                     wr             = 3,
                     dly            = 4,
                     rd             = 5;

   reg [4:0]         next_state;
   reg [4:0]         next_state1;
   reg [5:0]         init_dly;
   reg               lfsr_rst_180;
   reg               lfsr_rst_90;
   reg               init_done;
   reg               rst_flag;
   reg               temp;
   reg               rst90_r;
   reg               rst180_r;
   wire [5:0]        init_dly_p;
   wire [3:0]        u_cmd_p;
   wire              addr_inc_p;
   wire              lfsr_rst_p;
   wire              next_cmd;

   always @( posedge clk90 )
     rst90_r <= rst90;

   always @( negedge clk )
     rst180_r <= rst180;

   assign            lfsr_rst       = lfsr_rst_90;

   assign            u_cmd_p          = (next_state == rd )  ? 4'b0110 :
                     (next_state == wr)           ? 4'b0100 :
                     (next_state == init_start)   ? 4'b0010 :
                     4'b0000;

   assign            addr_inc_p     = ((cmd_ack == 1'b1) && (next_state == wr ||
                                                           next_state == rd));
   assign            addr_rst       = rst_flag ;

   always@(negedge clk) begin
      rst_flag <= ( !rst180_r && !cmd_ack && !temp);
      temp         <= ( !rst180_r && !cmd_ack);
   end

   assign lfsr_rst_p     =   (r_w == 1'b1) ? 1'b1 : 1'b0 ;
   assign init_dly_p     = (next_state == init_start) ? 6'b111111 :
          (init_dly != 6'b000000) ? init_dly - 1'b1 :
          6'b000000;


   assign next_cmd = ((cmd_ack == 1'b0) && next_state == dly);

   always @ (negedge clk) begin
      if (rst180_r == 1'b1)
        lfsr_rst_180  <= 1'b0;
      else
        lfsr_rst_180  <= lfsr_rst_p;
   end

   always @ (posedge clk90) begin
      if (rst90_r == 1'b1) begin
         lfsr_rst_90  <= 1'b0;
      end
      else begin
         lfsr_rst_90  <= lfsr_rst_180;

      end
   end

   always @ (negedge clk) begin
      if (rst180_r == 1'b1) begin
         u_cmd <= 4'b0000;
      end
      else begin
         u_cmd <= u_cmd_p;
      end
   end


   always @ (negedge clk) begin
      if (rst180_r == 1'b1) begin
         addr_inc  <= 1'b0;
         init_dly  <= 6'b000000;
      end
      else begin
         addr_inc  <= addr_inc_p;
         init_dly  <= init_dly_p;
      end
   end

   always @ (negedge clk) begin
      if (rst180_r == 1'b1)
        init_done <= 1'b0;
      else
        init_done <= init_val;
   end

   always@(negedge clk) begin
      if(rst180_r == 1'b1)
        r_w <= 1'b0;
      else
        if(cmd_ack == 1'b0 && next_state == rd )
          r_w <= 1'b1;
        else if(cmd_ack == 1'b0 && next_state == wr )
          r_w <= 1'b0;
        else
          r_w  <= r_w;
   end


      always @ (next_state or rst180_r or cnt_roll or r_w or refresh_done or 
	              init_done or next_cmd) begin
         if (rst180_r == 1'b1)
           next_state1 <= rst_state;
         else begin
            case (next_state)
              rst_state : begin
                 next_state1 <= init_start;
              end
              
              init_start : begin
                 next_state1 <= init;
              end
              
              init : begin
                 if (init_done == 1'b1)
                   next_state1 <= wr;
                 else
                   next_state1 <= init;
              end

              wr :  begin
                 if (cnt_roll == 1'b0)
                   next_state1 <= wr;
                 else
                   next_state1 <= dly;
              end
	      
              dly : begin
                 if(next_cmd == 1'b1 && r_w == 1'b0)
                   next_state1 <= rd;
                 else if(next_cmd == 1'b1 && r_w == 1'b1)
                next_state1 <= wr;
                 else
                   next_state1 <= dly;
              end
             rd : begin
                 if (cnt_roll == 1'b0)
                   next_state1 <= rd;
                 else
                   begin
                      next_state1 <= dly;
                   end
              end
              default : begin
                 next_state1 <= rst_state;
              end
            endcase
         end
      end

   always @ (negedge clk) begin
      if (rst180_r == 1'b1) begin
         next_state    <= rst_state;
      end
      else begin
         next_state    <= next_state1;
      end
   end

endmodule
