///////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005 Xilinx, Inc.
// This design is confidential and proprietary of Xilinx, All Rights Reserved.
///////////////////////////////////////////////////////////////////////////////
//   ____  ____
//  /   /\/   /
// /___/  \  /   Vendor			: Xilinx
// \   \   \/    Version		: $Name: mig_v1_7_b7 $
//  \   \        Application		: MIG
//  /   /        Filename		: vlog_xst_bl4_data_path_0.v
// /___/   /\    Date Last Modified	: $Date: 2007/02/20 10:35:45 $
// \   \  /  \   Date Created		: Mon May 2 2005
//  \___\/\___\
// Device	: Spartan-3/3A
// Design Name	: DDR2 SDRAM
// Purpose	: This module has the write and read data paths for the
//                DDR2 memory interface. The write data along with write enable
//                signals are forwarded to the DDR IOB FFs. The read data is
//                captured in CLB FFs and finally input to FIFOs.
///////////////////////////////////////////////////////////////////////////////

`include "../source/sdramtst/vlog_xst_bl4_parameters_0.v"
`timescale 1ns/100ps
module vlog_xst_bl4_data_path_0
  (
   input      [((`data_width*2)-1):0]  user_input_data,
   input [((`data_mask_width*2) -1):0] user_data_mask,
   input                               clk,
   input                               clk90,
   input                               reset,
   input                               reset90,
   input                               write_enable,
   input                               rst_dqs_div_in,
   input [4:0]                         delay_sel,
   input [(`data_width-1):0]           dq,
   input                              dqs_int_delay_in0,
   input                              dqs_int_delay_in1,
   output                              u_data_val,
   output [((`data_width*2)-1):0]      user_output_data,
   output                              write_en_val,
   output [((`data_mask_width)-1):0]   data_mask_f,
   output [((`data_mask_width)-1):0]   data_mask_r,
   output [(`data_width-1):0]          write_data_falling,
   output [(`data_width-1):0]          write_data_rising
   );



   wire                                reset_r;
   wire [3:0]                          fifo0_rd_addr;
   wire [3:0]                          fifo1_rd_addr;
   wire                                read_valid_data_1;
   wire [3:0]                          fifo_00_wr_addr;
   wire [3:0]                          fifo_01_wr_addr;
   wire [3:0]                          fifo_10_wr_addr;
   wire [3:0]                          fifo_11_wr_addr;
   wire                             fifo_00_wr_en;
   wire                             fifo_01_wr_en;
   wire                             fifo_10_wr_en;
   wire                             fifo_11_wr_en;
   wire                                dqs0_delayed_col0;
   wire                                dqs0_delayed_col1;
   wire                                dqs1_delayed_col0;
   wire                                dqs1_delayed_col1;

   vlog_xst_bl4_data_read_0 data_read0
     (
      .clk90               (clk90),
      .reset90_r           (reset90),
      .ddr_dq_in           (dq),
      .read_valid_data_1   (read_valid_data_1),
      .fifo_00_wr_en      ( fifo_00_wr_en),
      .fifo_01_wr_en      ( fifo_01_wr_en),

      .fifo_10_wr_en      ( fifo_10_wr_en),
      .fifo_11_wr_en      ( fifo_11_wr_en),

      .fifo_00_wr_addr     (fifo_00_wr_addr),
      .fifo_01_wr_addr     (fifo_01_wr_addr),
      .fifo_10_wr_addr     (fifo_10_wr_addr),
      .fifo_11_wr_addr     (fifo_11_wr_addr),
      .dqs0_delayed_col0   (dqs0_delayed_col0),
      .dqs0_delayed_col1   (dqs0_delayed_col1),
      .dqs1_delayed_col0   (dqs1_delayed_col0),
      .dqs1_delayed_col1   (dqs1_delayed_col1),
      .user_output_data    (user_output_data),
      .fifo0_rd_addr_val   (fifo0_rd_addr),
      .fifo1_rd_addr_val   (fifo1_rd_addr )
      );
   vlog_xst_bl4_data_read_controller_0  data_read_controller0
     (
      .clk90                  (clk90),
      .reset_r                (reset_r), 
      .reset90_r              (reset90),
      .rst_dqs_div_in         (rst_dqs_div_in),
      .delay_sel              (delay_sel),
      .dqs_int_delay_in0      (dqs_int_delay_in0),
      .dqs_int_delay_in1      (dqs_int_delay_in1),
      .fifo_00_wr_en_val (fifo_00_wr_en),
      .fifo_01_wr_en_val (fifo_01_wr_en),
      .fifo_10_wr_en_val (fifo_10_wr_en),
      .fifo_11_wr_en_val (fifo_11_wr_en),
      .fifo_00_wr_addr_val    (fifo_00_wr_addr),
      .fifo_01_wr_addr_val    (fifo_01_wr_addr),
      .fifo_10_wr_addr_val    (fifo_10_wr_addr),
      .fifo_11_wr_addr_val    (fifo_11_wr_addr),
      .dqs0_delayed_col0_val  (dqs0_delayed_col0),
      .dqs0_delayed_col1_val  (dqs0_delayed_col1),
      .dqs1_delayed_col0_val  (dqs1_delayed_col0),
      .dqs1_delayed_col1_val  (dqs1_delayed_col1),
      .fifo0_rd_addr          (fifo0_rd_addr),
      .fifo1_rd_addr          (fifo1_rd_addr),
      .u_data_val             (u_data_val),
      .read_valid_data_1_val  (read_valid_data_1)
      );


   vlog_xst_bl4_data_write_0 data_write0
     (
      .user_input_data    (user_input_data),
      .user_data_mask     (user_data_mask),
      .clk90              (clk90),
      .write_enable       (write_enable),
      .write_en_val       (write_en_val),
      .write_data_falling (write_data_falling),
      .write_data_rising  (write_data_rising),
      .data_mask_f        (data_mask_f),
      .data_mask_r        (data_mask_r)
      );

   vlog_xst_bl4_data_path_rst  data_path_rst0
     (
      .clk                (clk),
      .reset              (reset),
      .reset_r            (reset_r)
      );

endmodule
