///////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005 Xilinx, Inc.
// This design is confidential and proprietary of Xilinx, All Rights Reserved.
///////////////////////////////////////////////////////////////////////////////
//   ____  ____
//  /   /\/   /
// /___/  \  /   Vendor			: Xilinx
// \   \   \/    Version		: $Name: mig_v1_7_b7 $
//  \   \        Application		: MIG
//  /   /        Filename		:  vlog_xst_bl4_infrastructure_top.v
// /___/   /\    Date Last Modified	: $Date: 2007/02/19 12:03:14 $
// \   \  /  \   Date Created		: Mon May 2 2005
//  \___\/\___\
// Device	: Spartan-3/3A
// Design Name	: DDR2 SDRAM
// Purpose	: This module has instantiations clk_dcm and cal_top. 
//		  This generate reset signals.
///////////////////////////////////////////////////////////////////////////////

`timescale 1ns/100ps
`include "../source/sdramtst/vlog_xst_bl4_parameters_0.v"
module vlog_xst_bl4_infrastructure_top
  (
   input         SYS_CLK,
   input         reset_in_n,
   output [4:0]  delay_sel_val1_val,
   output        sys_rst_val,
   output        sys_rst90_val,
   output        sys_rst180_val,
   output reg    wait_200us_rout,
   output        clk_int_val,
   output        ckdiv_out,
   output        clocklock,
   output        clk90_int_val
   );

   wire          user_rst;
   wire          clk_int;
   wire          clk90_int;
   wire          dcm_lock;
   wire          sys_clk_ibuf;
   wire          clk_int_val1;
   wire          clk_int_val2;
   wire          clk90_int_val1;
   wire          clk90_int_val2;
   wire [4:0]    delay_sel_val;
   wire          user_cal_rst;
   reg           sys_rst_o;
   reg           sys_rst_1;
   reg           sys_rst;
   reg           sys_rst90_o;
   reg           sys_rst90_1;
   reg           sys_rst90;
   reg           sys_rst180_o;
   reg           sys_rst180_1;
   reg           sys_rst180;
   reg [15:0]    Counter200;
   reg           wait_200us;
   reg           wait_clk90;
   reg           wait_clk270;
   reg           wait_200us_r; 

//   IBUFG  lvds_clk_input
//     (
//      .I(SYS_CLK),
//      .O(sys_clk_ibuf)
//      );

   assign sys_clk_ibuf = SYS_CLK;

      
   assign       clk_int_val    = clk_int;
   assign       clk90_int_val  = clk90_int;
   assign       sys_rst_val    = sys_rst;
   assign       sys_rst90_val  = sys_rst90;
   assign       sys_rst180_val = sys_rst180;
   assign       delay_sel_val1_val = delay_sel_val;


//To remove delta delays in the clock signals observed during simulation 
// ,Following signals are used

   assign       clk_int_val1   = clk_int;
   assign       clk90_int_val1 = clk90_int;
   assign       clk_int_val2   = clk_int_val1;
   assign       clk90_int_val2 = clk90_int_val1;
   assign       user_rst       = `reset_active_low  == 1'b1 ? ~reset_in_n :  reset_in_n;
   assign       user_cal_rst   = `reset_active_low  == 1'b1 ?  reset_in_n : ~reset_in_n;

   always @(posedge clk_int_val2) begin
      if(user_rst == 1'b1 || dcm_lock == 1'b0) begin
         wait_200us <= 1'b1;
         Counter200     <= 16'd0;
      end
      else
`ifdef simulation
        wait_200us <= 1'b0;
`else
     // if ( Counter200 < 16'd33200 ) begin
      if ( Counter200 < 16'd44000 ) begin
         Counter200 <= Counter200 + 1'b1;
         wait_200us <= 1'b1;
      end
      else begin
         Counter200 <= Counter200;
         wait_200us <= 1'b0;
      end
`endif
   end
   
   always @( posedge clk_int_val2 )
     wait_200us_r <= wait_200us;

   always @( posedge clk_int_val2 )
     wait_200us_rout <= wait_200us;

   always @(negedge clk90_int_val2) begin
      if(user_rst == 1'b1 || dcm_lock == 1'b0)
        wait_clk270 <= 1'b1;
      else
        wait_clk270 <= wait_200us_r;
   end

   always @(posedge clk90_int_val2) begin
      wait_clk90 <= wait_clk270;
   end
   
   always@(posedge clk_int_val2) begin
      if(user_rst == 1'b1 || dcm_lock == 1'b0 || wait_200us_r == 1'b1 ) begin
         sys_rst_o <= 1'b1;
         sys_rst_1 <= 1'b1;
         sys_rst   <= 1'b1;
      end
      else begin
         sys_rst_o <= 1'b0;
         sys_rst_1 <= sys_rst_o;
         sys_rst   <= sys_rst_1;
      end
   end

   always@(posedge clk90_int_val2) begin
      if (user_rst == 1'b1 || dcm_lock == 1'b0 || wait_clk90 == 1'b1) begin
         sys_rst90_o <= 1'b1;
         sys_rst90_1 <= 1'b1;
         sys_rst90   <= 1'b1;
      end
      else begin
         sys_rst90_o <= 1'b0;
         sys_rst90_1 <= sys_rst90_o;
         sys_rst90   <= sys_rst90_1;
      end
   end

   always@(negedge clk_int_val2) begin
      if (user_rst == 1'b1 || dcm_lock == 1'b0 || wait_clk270 == 1'b1) begin
         sys_rst180_o <= 1'b1;
         sys_rst180_1 <= 1'b1;
         sys_rst180   <= 1'b1;
      end
      else begin
         sys_rst180_o <= 1'b0;
         sys_rst180_1 <= sys_rst180_o;
         sys_rst180   <= sys_rst180_1;
      end
   end

   assign clocklock = dcm_lock;

   vlog_xst_bl4_clk_dcm clk_dcm0
     (
      .input_clk   ( sys_clk_ibuf),
      .rst         ( user_rst),
      .clk         ( clk_int),
      .clk90       ( clk90_int),
      .ckdiv_out   ( ckdiv_out),
      .dcm_lock    ( dcm_lock)
      );

   vlog_xst_bl4_cal_top cal_top0
     (
      .clk0        ( clk_int_val2),
      .clk0dcmlock ( dcm_lock),
      .reset       ( user_cal_rst),
      .tapForDqs   ( delay_sel_val)
      );

endmodule
