///////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2005 Xilinx, Inc.
// This design is confidential and proprietary of Xilinx, All Rights Reserved.
///////////////////////////////////////////////////////////////////////////////
//   ____  ____
//  /   /\/   /
// /___/  \  /   Vendor			: Xilinx
// \   \   \/    Version		: $Name: mig_v1_7_b7 $
//  \   \        Application		: MIG
//  /   /        Filename		: vlog_xst_bl4_top_0.v
// /___/   /\    Date Last Modified	: $Date: 2007/02/20 10:35:45 $
// \   \  /  \   Date Created		: Mon May 2 2005
//  \___\/\___\
// Device	: Spartan-3/3A
// Design Name	: DDR2 SDRAM
// Purpose	: This modules has the instantiations infrastructure, iobs, 
//		  controller and data_paths modules
///////////////////////////////////////////////////////////////////////////////

`timescale 1ns/100ps
`include "../source/sdramtst/vlog_xst_bl4_parameters_0.v"

module  vlog_xst_bl4_top_0
  (
   input                                         rst_dqs_div_in,
   output                                        rst_dqs_div_out,
   input                                         clk_int,
   input                                         clk90_int,
   input [4:0]                                   delay_sel_val,
   input                                         refr_enable,
   input                                         sys_rst,
   input                                         sys_rst90,
   input                                         sys_rst180,
   input [((`data_width*2)-1):0]                 user_input_data,
   output [((`data_width*2)-1):0]                user_output_data,
   output                                        user_data_valid,
   input [((`row_address +
            `col_ap_width + `bank_address)-1):0] user_input_address,
   input [3:0]                                   user_command_register,
   output                                        user_cmd_ack,
   input                                         burst_done,
   output                                        init_val,
   output                                        ar_done,
   inout [((`data_strobe_width)-1):0]            DDR2_DQS,
   inout [(`data_width-1):0]                     DDR2_DQ,
   output                                        DDR2_CKE,
   output                                        DDR2_CS_N,
   output                                        auto_ref_req,
   input                                         wait_200us,
   output                                        DDR2_RAS_N,
   output                                        DDR2_CAS_N,
   output                                        DDR2_WE_N,
   output                                        DDR2_ODT,
   output [`bank_address-1:0]                    DDR2_BA,
   output [`row_address-1:0]                     DDR2_A,
   output [(`clk_width-1):0]                     DDR2_CK,
   output       [((`data_mask_width)-1):0]  DDR2_DM,
   input      [((`data_mask_width*2)-1):0] user_data_mask,
    inout  [(`data_strobe_width-1):0] DDR2_DQS_N, 


   output [(`clk_width-1):0]                     DDR2_CK_N
   );


   wire                                          rst_calib;
   wire [4:0]                                    delay_sel;
   wire                                          write_enable;
   wire                                          dqs_div_rst;
   wire                                          dqs_enable;
   wire                                          dqs_reset;
   wire [(`data_width-1):0]                      dq;
   wire                                          write_en_val;
   wire                                          reset90_r;
   wire [((`data_mask_width)-1):0]               data_mask_f;
   wire [((`data_mask_width)-1):0]               data_mask_r;
   wire [(`data_width-1):0]                      write_data_falling;
   wire [(`data_width-1):0]                      write_data_rising;
   wire                                          ddr_rasb_cntrl;
   wire                                          ddr_casb_cntrl;
   wire                                          ddr_web_cntrl;
   wire [`bank_address-1:0]                      ddr_ba_cntrl;
   wire [`row_address-1:0]                       ddr_address_cntrl;
   wire                                          ddr_cke_cntrl;
   wire                                          ddr_csb_cntrl;
   wire                                          ddr_ODT_cntrl;
   wire                                          rst_dqs_div_int;
   wire	[((`data_mask_width*2)-1):0]		 user_mask;

   wire dqs_int_delay_in0; 
   wire dqs_int_delay_in1; 


	  assign  user_mask	=	user_data_mask;

   vlog_xst_bl4_controller_0 controller0
     (
      .refr_enable       (refr_enable),
      .auto_ref_req      (auto_ref_req),
      .wait_200us(wait_200us),
      .clk               (clk_int),
      .rst0              (sys_rst),
      .rst180            (sys_rst180),
      .address           (user_input_address[((`row_address + 
                                               `col_ap_width +
                                               `bank_address)-1):
                                             `bank_address]),
      .bank_address      (user_input_address[(`bank_address-1): 0]),
      .command_register  (user_command_register),
      .burst_done        (burst_done),
      .ddr_rasb_cntrl    (ddr_rasb_cntrl),
      .ddr_casb_cntrl    (ddr_casb_cntrl),
      .ddr_web_cntrl     (ddr_web_cntrl),
      .ddr_ba_cntrl      (ddr_ba_cntrl),
      .ddr_address_cntrl (ddr_address_cntrl),
      .ddr_cke_cntrl     (ddr_cke_cntrl),
      .ddr_csb_cntrl     (ddr_csb_cntrl),
      .ddr_ODT_cntrl     (ddr_ODT_cntrl),
      .dqs_enable        (dqs_enable),
      .dqs_reset         (dqs_reset),
      .write_enable      (write_enable),
      .rst_calib         (rst_calib),
      .rst_dqs_div_int   (rst_dqs_div_int),
      .cmd_ack           (user_cmd_ack),
      .init              (init_val),
      .ar_done           (ar_done)
      );
   
   vlog_xst_bl4_data_path_0  data_path0
     (
      .user_input_data    (user_input_data),
      .user_data_mask    (user_mask),
      .clk               (clk_int),
      .clk90             (clk90_int),
      .reset              (sys_rst),
      .reset90            (sys_rst90),
      .write_enable       (write_enable),
      .rst_dqs_div_in     (dqs_div_rst),
      .delay_sel          (delay_sel),
      .dqs_int_delay_in0  (dqs_int_delay_in0),
      .dqs_int_delay_in1  (dqs_int_delay_in1),
      .dq                 (dq),
      .u_data_val         (user_data_valid),
      .user_output_data   (user_output_data),
      .write_en_val       (write_en_val),
      .data_mask_f        (data_mask_f),
      .data_mask_r        (data_mask_r),
      .write_data_falling (write_data_falling),
      .write_data_rising  (write_data_rising)
      );
   
   vlog_xst_bl4_infrastructure infrastructure0
     (
      .clk_int(clk_int),
      .rst_calib1(rst_calib),
      .delay_sel_val(delay_sel_val),
      .delay_sel_val1_val(delay_sel)
      );
   
   vlog_xst_bl4_iobs_0       iobs0
     (
      .clk               (clk_int),
      .clk90            (clk90_int),
      .ddr_rasb_cntrl    (ddr_rasb_cntrl),
      .ddr_casb_cntrl    (ddr_casb_cntrl),
      .ddr_web_cntrl     (ddr_web_cntrl),
      .ddr_cke_cntrl     (ddr_cke_cntrl),
      .ddr_csb_cntrl     (ddr_csb_cntrl),
      .ddr_ODT_cntrl     (ddr_ODT_cntrl),
      .ddr_address_cntrl (ddr_address_cntrl),
      .ddr_ba_cntrl      (ddr_ba_cntrl),
      .rst_dqs_div_int   (rst_dqs_div_int),
      .dqs_reset         (dqs_reset),
      .dqs_enable        (dqs_enable),
      .ddr_dqs           (DDR2_DQS),
       .ddr_dqs_n (DDR2_DQS_N),
      .ddr_dq            (DDR2_DQ),
      .write_data_falling(write_data_falling),
      .write_data_rising (write_data_rising),
      .write_en_val      (write_en_val),
      .data_mask_f       (data_mask_f),
      .data_mask_r       (data_mask_r),
      .DDR2_CK             (DDR2_CK),
      .DDR2_CK_N             (DDR2_CK_N),
      .ddr_rasb          (DDR2_RAS_N),
      .ddr_casb          (DDR2_CAS_N),
      .ddr_web           (DDR2_WE_N),
      .ddr_ba            (DDR2_BA),
      .ddr_address       (DDR2_A),
      .ddr_cke           (DDR2_CKE),
      .ddr_csb           (DDR2_CS_N),
      .ddr_ODT0          (DDR2_ODT),
      .rst_dqs_div       (dqs_div_rst),
      .rst_dqs_div_in  (rst_dqs_div_in),
      .rst_dqs_div_out         (rst_dqs_div_out),
      .dqs_int_delay_in0  (dqs_int_delay_in0),
      .dqs_int_delay_in1  (dqs_int_delay_in1),
      .ddr_dm          (DDR2_DM),
      .dq                (dq)
      );
   
endmodule
